#! /usr/bin/perl -w

use strict;

# 	Invocation parameters:
#	1. File(s) containing CDG2 annotation from which to extract the thematic roles assignments.


my $old_word_number = 2;
my $current_word_number;
my $word;
my @sentence;

my $label;
my %label_to_position_attached;
# 	This hash contains a map of edge label upon the word in the sentence which is attached via that label. Note that this may cause problems
# 	when a label occurs multiple times in a sentence.
my %direct_attachments;
my %all_attachments;
# 	This hash contains a map of words upon the list of words attaching to that word (each word is represented by its position in the sentence)
my $line;
my %label_of_level_by_position; # Saves the label of a word by every level and position

my %print_levels = ( "AGNT" => 0, "THME" => 0, "RCPN" => 0, "INST" => 0); #these levels should be printed

my %labels_to_level =	( 	
					"SUBJ"	=> 	"SYN",
					"OBJA"		=>	"SYN",
				);	
# 					"AGENT"		=>	"AGNT",
# 					"INSTRUMENT"	=>	"INST",
# 					"THEME"		=>	"THME",
# 					"COMITATIVE"	=>	"INST",
# 					"RECIPIENT"		=>	"RCPN",
# 				);
		
#		my @array = %labels;
# 		print "@array";
# print "\n\n annotations in $ARGV[0]";
# print "\n\n sentences in $ARGV[1]";

	# Initialise the hash values for the labels_to_level to ensure the output contains the same number
	# of lines for every parse. 
	for my $label ( keys %labels_to_level )
	{
		$label_to_position_attached{ $label } = "-";
	}
	
	while ( <> )
	{	
		#print "================================\n\n" if ( /====/ );		# If the next line is a separator it is also the end of a block of sentences.
		
		if ( /\s+[0-9]+ (\d+) ([\w',.:;:!?]+)$/io ) 	# If the line contains the word, extract word and word number in sentence.
		{	
			$current_word_number = $1;			
			$word = $2;
			$old_word_number = $current_word_number;
			push @sentence, $word;
		}
		elsif( /^;$/ )						# If the line marks the end of the sentence (";" in the annotation file), print out all findings for this sentence and reset.
		{
		
			print join " ", @sentence, "\n";
			
			for my $position ( 1 .. scalar @sentence )		# labels_to_level are listed in inverse ascibetical order.
			{	
				my @temp = ();
				if(get_all_attachments( $position))
				{
					@temp = @{get_all_attachments( $position)};
					# push @temp, $position;
				}
# 				if(@temp)
# 				{

					push @temp, $position;

# 				}

				#removes non-unique elements from array
				my %all=();
				@all{@temp}=1;
				@temp = (keys %all);


				#get_all_attachments( $position);
				 #printf "%-15s %-15s\n", $position, @{ &get_all_attachments( $position) };
				#printf "%-15s %-15s\n", $position, ( $direct_attachments{ $position } ? "@{ $direct_attachments{ $position } }" : "" );
				#printf "%-15s %-15s\n", $position, ( $all_attachments{ $position } ? "@{ $all_attachments{ $position } }" : "" );
				#printf "%-15s %-15s\n", $position, ( get_all_attachments( $position) ? "@{ get_all_attachments( $position) }" : "" );


				#print "$position\t";

				foreach my $level (keys %print_levels)

				#foreach my $level (keys %{$label_of_level_by_position{$position}})
				{

				#print $level;

					if($label_of_level_by_position{$position}{$level})
					{	

						print "$label_of_level_by_position{$position}{$level}\t";

# 						

						for my $attachment ( sort{$a<=>$b} (@temp)  )
						{

							
			
							print "$sentence[$attachment-1] ";
		
						}
		
						print "\n";

					}

				}

			}
			
			print "\n";
			
			
			# Now reset all sentence-specific data structures.
			
			@sentence = ();
			
			for $label ( keys %labels_to_level )
			{
				$label_to_position_attached{ $label } = "-";
			}
			
			%direct_attachments = ();
			%label_of_level_by_position = ();
			
		}
		elsif( /(\w{3,4}) -> (\w+) -> ([0-9]+)/io ) 		# If the line contains a word-to-word attachment, extract word attached via that label.
		{	
			push @{ $direct_attachments{ $3 } }, $current_word_number;	# Position $2 has position $current_word_number as dependant.

			$label_of_level_by_position{$current_word_number}{$1} = $2; # Save the label of a word by every level and position	

			
			for $label ( keys %labels_to_level )			# Loop over all labels.
			{		
				if ( /$labels_to_level{ $label } -> ($label) -> ([0-9]+)/ ) 		# Check for match, e. g. in the line "SYN -> SUBJ -> 5".
				{
					$label_to_position_attached{ $2 } = $current_word_number  unless ( $2 eq '\'\'' );
					last;
				}
			}
		}
	}

	
	sub get_all_attachments
	{
		my $position = shift @_;
		my %all_attachments = ();

		if($direct_attachments{ $position })
		{
			push @{ $all_attachments{ $position } }, @{$direct_attachments{ $position }} ;
			
		}

		
		
		for my $attachment ( @{ $direct_attachments{ $position } } )
		{	
			
			my $temp = get_all_attachments( $attachment );
			
			if($temp)
			{
				push @{ $all_attachments{ $position } }, @{$temp} ;
				
			}
		}
		
	
		
		return $all_attachments{ $position };
	}

	#print "================================\n\n";