#! /usr/bin/perl -w

use strict;
use locale;

#	This script extracts sentence annotations from a CDG annotation file containing
#	concatenated sentence annotations for multiple sentences. Extraction is based
#	on sentence number. The numbers of the sentences whose annotation is to be extacted 
#	are either obtained from one or more input files or can be a simple list of numbers 
#	handed over as a command line parameter.
#
#	Note that in the current implementation the sentence numbers in the extracted annotations 
#	are automatically renumbered in the output to match the new sentence counts.
#
#	Invocation Arguments
#	====================
#	1. CDG annotation file from which to extract the sentence annotations.
#	2. File(s) containing the sentence numbers to extract or sentence numbers to extract.


my %extract;

open ANNOTATIONS, "<", shift @ARGV;

for my $argument ( @ARGV )	# Loop over all invocation arguments.
{
	if ( -f $argument ) # If the invocation argument is a file, read in the numbers from the file.
	{
		open EXTRACT, "<", $argument or die "Could not open the file $argument.";
		
		$extract{ eval $_ } ++ for <EXTRACT>;
	}
	else 	# If the invocation argument is not a file, treat the argument as a sentence number.
	{
			$extract{ $argument } ++;	
	}
}

# print "There are annotations for", scalar keys %extract, " sentences to extract.\n";

my $counter = 0;
my $printer = 0;

while ( <ANNOTATIONS> )
{
   $counter ++ if /wordgraph/o;
   $printer ++ if ( /wordgraph/o && $extract{ $counter } );
   
   my $string = sprintf "%03d", $printer;	# Make '$string' a three digit numeral with leading zeros is required.
   s/\d{3,3}/$string/;						# Substitute the old sentence number with the new one.
   print if ( $extract{ $counter } );
} 
