#! /usr/bin/perl -w

use strict;
use locale;

#	This script merges two CDG annotation files, e.g.: one file containing the correct sentences of a gold standard with  
#	a second file containing the corrected annotations for the gold standard into a single correct gold standard file.
#	The sentence numbers of the sentences that were corrected are obtained from an input file.
#	Note that due to the recognition of annotations based on an annotation header line the algorithm is actually 
#	more involved than initially meets the eye!
#
#	Invocation Arguments
#	====================
#	1. The preliminary gold standard file from which the incorrect CDG annotations have been removed.
#	2. The file containing the corrected CDG annotations.
#	3. The file containing the sentence numbers of the sentences that have been corrected.


my %insert;

open GOLD, "<", shift @ARGV or die "Could not open the gold standard file.";
open CORRECTIONS, "<", shift @ARGV or die "Could not open the corrections file.";
open NUMBERS, "<", shift @ARGV or die "Could not open the sentence number file.";


while ( <NUMBERS> )
{
	chomp;
	$insert{ $_ } ++ ;
}

# print "There are ", scalar keys %insert, " sentences to insert into the preliminary gold standard.\n";

my $sentence_counter = 0;

while ( <GOLD> )	# Run through the preliminary gold standard file and insert corrected sentences were necessary.
{
	my $line = $_;
     	$sentence_counter ++ if /wordgraph/o;	# Detect a new annotation based on its header line.
     	
 	if ( $insert{ $sentence_counter } )			# Check if an insertion has been defined for this sentence.
	{	
		while ( $insert{ $sentence_counter } )		# Loop over the number of insertions in this place.
		{
			while ( <CORRECTIONS> )		# Insert all lines for a single corrected sentence.
			{
				chomp;
				print $_, "\n"; 
				last if ( /;/ );
			}
		
			$sentence_counter ++;			# Increase the sentence count after each insertion.
		}
	}
	print $line;
}
