package cia_components.ppc;

import java.io.*;
import java.util.*;
import java.net.*;

public class Sentence
{
	Vector <String> input_from_cdg_main;
	Vector <String> sentence_word_strings;
	Vector <Slot> sentence_slots;
	Vector <Word> sentence_words;
	Vector <String> input_from_cdg_predictor;
	Vector <String> relevant_input_from_cdg_predictor;
	Vector <String> homonym_descriptions;
	
	// HashMap hypothesised_attachments;
	String ppc_integration_mode;
	
	int sentence_number;
	int length;
		
	public Sentence( int id )
	{	
		set_Sentence_Number( id );
		
		set_Length( -1 );
		// Sentence length starts at -1 to compensate for the first line received from cdg which is discarded.
		
		/*
			Depending on whether the PPC is to be run in homonym-specific or -unspecific predictor mode 
			input from CDG will be received either via a dynamically generated temp file or via STDIN.
		*/
		
		homonym_descriptions = new Vector <String> ();
		sentence_words = new Vector <Word> ();
		
		ppc_integration_mode = System.getenv( "PPC_INTEGRATION_MODE" );
		
		if( ppc_integration_mode.equals( "1" ) | ppc_integration_mode.equals( "2" ) | ppc_integration_mode.equals( "3" ) )
		{
			read_Input_From_CDG_Main_To_PPC_Via_File();
		}
		else if( ppc_integration_mode.equals( "4" ) )
		{
			read_Input_From_CDG_Main_Via_STDIN();
			// System.out.println( "PPC Integration Mode is: 4" );
		}

		make_Sentence_Slots();
		
		get_Descriptions();
			
		/*  
			This code may be of use when integrating input from CDG as its own predictor:
			read_Input_From_CDG_Predictor();
			hypothesised_attachments = new HashMap();
			// This hashmap maps a sentence slot (representing single word or the head of a phrase) onto the 
			// sentence slot it is hypothesised by the CDG predictor to attach to.
			relevant_input_from_cdg_predictor = new Vector();
			set_Attachments_Predicted_By_CDG();
		*/
	}
	
		
	public void read_Input_From_CDG_Main_To_PPC_Via_File()
	{
		/*
			This method reads all input lines obtained from the CDG main programme via the temporary file that CDG generates for
			communication with a registered predictor. The name of the temporary file is handed over as a command line argument 
			to the PPC starter script 'Start_PPC' in the directory /DATE/cia_components/.
		*/
			
		input_from_cdg_main = new Vector <String> ();
		// This vector holds the untokenized input lines received from the main instance of cdg.
								
		String cdg_to_ppc_file = "";
		String line = "";
		
		try
		{
			cdg_to_ppc_file = System.getenv( "CDG_MAIN_TO_PPC_FILE" );
			// System.out.println( "File containing the input from CDG: " + cdg_to_ppc_file );
			
			FileReader file_reader = new FileReader( cdg_to_ppc_file );
			BufferedReader buffered_reader = new BufferedReader( file_reader );
			
					
			while ( ( line = buffered_reader.readLine() ) != null ) 
			{
				line = line.trim();
				// System.out.println( "Line received from CDG: \t" + line );
				input_from_cdg_main.add( line );
			}	
			file_reader.close(); 
		}
		catch (IOException e)
		{
			System.err.println( "Could not read file " + cdg_to_ppc_file + " because: \n" + e.getMessage() + "\n" );
		}
	}

	
	public void read_Input_From_CDG_Main_Via_STDIN()
	{
		/*
			This method reads in the input lines obtained from the CDG main programme via Java's System.in, i. e.: the STDIN.
		*/
		
		input_from_cdg_main = new Vector <String> ();
		// This vector holds the untokenized input lines received from the main instance of cdg.
		
		InputStreamReader input_stream = new InputStreamReader( System.in );
		BufferedReader buffered_reader = new BufferedReader( input_stream );
						
		String line;
	
		// Now read in the input lines as received from CDG Main line-by-line via a buffered reader.
		try
		{ 		
			while ( ( line = buffered_reader.readLine() ) != null )
			{	
				// System.out.println( "Line received from CDG: \t" + line );
				input_from_cdg_main.add( line );
				input_from_cdg_main.add( "" );
				// This blank line is added to function as separator between the individual lines received from CDG.
				// It is included to permit the same handling for this input as for PPC input received from CDG via temp file.
			}		
		}
		catch ( Exception e )
		{
			System.err.println( "Could not read buffered input line from CDG because: \n" + e + "\n" );
		}
	}
		
		
	private void make_Sentence_Slots()
	{	
		/*
			This method creates all Slot objects an adds them to the vector 'sentence_slots'.
			The slot with index = 0 contains the ID line transmitted by cdg. 
			The full-stop's slot is the last slot of the sentence.
		*/
		
		String line;
		Vector <String> block = new Vector <String> ();
		Iterator <String> iter = input_from_cdg_main.iterator();
		
		sentence_slots = new Vector <Slot> ();
		// This vector holds all Slot objects for this sentence. It will be filled when generating a Slot object in Slot.class.
		
		while ( iter.hasNext() ) 		// Building the homonym blocks.
		{
			block.clear();	
			do
			{
				line = iter.next();
				block.add( line );
				// System.out.println( "Line: " + line );
			}
			while ( ! line.equals( "" ) );		// The end of a homonym block is marked by a blank separator line.
			
			block.removeElement( block.lastElement() );
			// This removes the last line from the vector 'block' because it always is just a blank separator line.
					
			// System.out.println( "Requesting slot based on block size: " + block.size() );
			Slot slot = new Slot( this, block );
			// Make the corresponding slot object based on the homonym block input.			
			
			sentence_slots.add( slot );
			// System.out.println( " Number of slots in sentence: " + sentence_slots.size() );
		}
		
		set_Length( sentence_slots.size() - 1 );
		// Set the length, i. e.: the number of slots, for this sentence. The correction of -1 is added 
		// because the ID slot is counted towards the size of the vector 'sentence_slots', too. 
		
		
		/*
			If input mode is 'classic' the separator lines which were inserted for processing purposes only 
			into the vector 'input_from_cdg_main' need to be removed again.
		*/	
		if( ppc_integration_mode.equals( "4" ) )
		{
			iter = input_from_cdg_main.iterator();
			
			while ( iter.hasNext() )
			{
				line = (String) iter.next();	
				if ( line.equals( "" ) )
				{
					iter.remove();
				}
			}
		}
	}


	private void retrieve_Homonym_Block_From_Script( String word )
	{	
		/*
			This method is only needed when wanting to retrieve homonyms for CDG1 predictor invocations. It is not needed when registering
			the PPC in CDG2 via 'predict -ppc ...' since this registration will ensure that all homonyms are provided by CDG.
			The method calls a bash script to actively retrieve homonyms from the lexicon based on a single word as input. From the script's 
			input it creates the slot's vector of homonym lines. Each vector element is a line containing the lexicon entry of one homonym. 
			Note that this method fails to retrieve homonyms which are generated by CDG's matches with unknown word templates.
		*/	
		Vector homonym_block = new Vector();
		String query;
		String slot_string = word;
		
		/*			
			The full-stop needs to be backslash-escaped as \. to prevent producing a list 
			of all single character entries in the lexicon. Without backslash-escape '.' would 
			be interpreted as 'arbitrary character' in the regular expression syntax. The second
			backslash is for Java to backslash-escape the backslash.
		*/
			
		if ( slot_string.equals( "." ) )
		{
			query = "\\.";	
		}
		else
		{
			query = slot_string;
		}
			
		Runtime run_time = Runtime.getRuntime();
		// This gets a new runtime instance. 		
		
		// Start the bash script which writes its output into a file.
		try 
		{
			String [] call = { System.getenv( "HOMONYM_RETRIEVER" ), query }; 
			Process call_script = run_time.exec( call );
			// This starts a new process in the runtime by executing the arguments given in the array.
			
			try
			{
				call_script.waitFor();
			}
			catch( InterruptedException e )
			{
				call_script.destroy();
				System.err.println( "Terminated the process because: \n" + e.getMessage() );
			}
		} 
		catch ( IOException e ) 
		{
			System.err.println( "Could not call script because: \n" + e.getMessage() );
		}
		
		// Now read in the output the script has written into its output file.
				
		String line = "";
		String homonym_file = "";
		try
		{
			homonym_file = System.getenv( "HOMONYM_TEMP_FILE" );
			FileReader file_reader = new FileReader( homonym_file );
			BufferedReader buffered_reader = new BufferedReader( file_reader );
					
			while ( ( line = buffered_reader.readLine() ) != null ) 
			{
				homonym_block.add( line );
				// Each line contains the literal lexicon entry for a given homonym.
				
				// System.out.println( "Homonym Line Found: " + line );
			}
		
			file_reader.close(); 
		}
		catch (IOException e)
		{
			System.err.println( "Could not read file " + homonym_file + " because: \n" + e.getMessage() );
		}
   	}

	
	private void read_Input_From_CDG_Predictor()
	{
		/*
			This method reads an annotation file produced by a CDG predictor instance. The predictor instance is a 
			non-contextualised CDG instance used to provide a structural prediction to the main instance of CDG for 
			how attachment would be without additional contextual knowledge. In the current implementation this 
			methods reads the entire file and assumes that each attachment found in the file corresponds to a word.
			
			Improvements needed to the method:
			(1) For sentences containing higher structures such as PPs additional attachments must be followed.
			A challenge will be the mapping of attachments to actual words in the sentence.
		*/
		
		input_from_cdg_predictor = new Vector();
		// This vector holds all input lines from the CDG prdictor's annotation file.
		File annotation = new File( System.getenv( "CDG_PREDICTOR_ANNOTATION" ) );
		// Convert the relative path into an absolute path.
		
		URI annotation_uri = annotation.toURI();
		String file_name = ( annotation_uri.normalize() ).toString();
		
		// System.out.println( file_name );
		
		// Remove the prefix 'file:' from the absolute path string.
		file_name = file_name.replace( "file:", "" );
	    
	    String line = "";
		
		// Now read in the annotation file line by line.
        try
        {
            FileReader file_reader = new FileReader( file_name );
            BufferedReader buffered_reader = new BufferedReader( file_reader );
                    
            while ( ( line = buffered_reader.readLine() ) != null )
            {
				input_from_cdg_predictor.add( line );
				// System.out.println( "Line read: " + line );
            }
			file_reader.close();
		}
        catch (IOException e)
        {
            System.err.println( "Opening the file " + file_name + " failed because:\n" + e.getMessage() );
        }       
	}
	
	
	private void set_Attachments_Predicted_By_CDG()
	{
		/*
			This method extracts the hypothesised attachment information from the annotation file
			produced by CDG's predictor instance.
		*/
		
		/*
		Iterator iter = input_from_cdg_predictor.iterator(); 
		String line;
		String extracted_word_number = "";
		String hypothesis = "";
		
		while ( iter.hasNext() )
		{
			line = (String) iter.next();
			
			// Extract which word is being hypothesised about.
			Pattern pattern_1 = Pattern.compile( "(\\d+\\s\\d+)" );
			Matcher matcher_1 = pattern_1.matcher( line );
			if ( matcher_1.find() )
			{
				line = line.trim();			
				extracted_word_number = line.substring( line.indexOf( ' ' ), line.lastIndexOf( ' ' ) ).trim();
				relevant_input_from_cdg_predictor.add( extracted_word_number + "," + extracted_word_number.length() );
			}
			
			// Extract which syntactic attachment is being hypothesised for that word.
			Pattern pattern_2 = Pattern.compile( "(->\\s\\d+)" );
			Matcher matcher_2 = pattern_2.matcher( line );
			
			if ( line.contains( "_SYN" ) & matcher_2.find() )
			{
				hypothesis = matcher_2.group().replace( "-> ", "" );
				relevant_input_from_cdg_predictor.add( hypothesis + "," + hypothesis.length() );
				hypothesised_attachments.put( get_Word_Object( Integer.valueOf( extracted_word_number )).get_Word_String(), Integer.valueOf( hypothesis ) );
			}
		}
		*/
	}
	

		
	public int get_Length()
	{
		return length;
	}
	
	
	public Slot get_Slot_Object( int i )
	{
		// System.out.println( "Requesting slot no. " + i );
		Slot result = null;
		result = (Slot) sentence_slots.elementAt( i );

		return result;
	}
	
	
	public Collection <Slot> get_All_Slot_Objects()
	{
		Collection <Slot> result = sentence_slots;
		return result;
	}


	public String get_Slot_String( int i )
	{
		String result = "";
		result =  (String) get_Slot_Object( i ).get_Slot_String();

		return result;
	}
	
	
	public Vector get_Input_From_CDG_Main()
	{
		return input_from_cdg_main;
	}
	
	
	public Vector get_Slots()
	{
		return sentence_slots;
	}
	
	
	public void add_Sentence_Word_String( String word )
	{
		sentence_word_strings.add( word );
	}
	
	
	public Vector get_Input_from_CDG_Predictor()
	{
		return input_from_cdg_predictor;
	}
	
	
//	public HashMap get_Hypothesised_Attachments()
//	{
//		return hypothesised_attachments;
//	}
	
	
	public void set_Sentence_Number( int i )
	{
		sentence_number = i;
	}
	
	
	public void set_Length( int i )
	{
		length = i;
	}
	
	
	public void add_Description( String input_string )
	{
		homonym_descriptions.addElement( input_string );
		// System.out.println( "Added description string '" + input_string + "' at sentence level." );
	}
	
	
	public Vector <String> get_Descriptions()
	{
		// System.out.println( homonym_description_strings.toString() );
		return homonym_descriptions;
	}
	
	
	public void add_Sentence_Word( Word word )
	{
		sentence_words.add( word );		
		// System.out.println( "Added word '" + word.get_Description() + "' to sentence." );
	}
	
	
	public Vector <Word> get_All_Word_Objects()
	{
		return sentence_words;		
	}
}