package cia_components.reasoning;

import java.net.URI;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.semanticweb.owl.apibinding.OWLManager;
import org.semanticweb.owl.inference.OWLReasoner;
import org.semanticweb.owl.inference.OWLReasonerAdapter;
import org.semanticweb.owl.inference.OWLReasonerException;
import org.semanticweb.owl.inference.OWLReasonerFactory;
import org.semanticweb.owl.model.OWLClass;
import org.semanticweb.owl.model.OWLDescription;
import org.semanticweb.owl.model.OWLIndividual;
import org.semanticweb.owl.model.OWLIndividualAxiom;
import org.semanticweb.owl.model.OWLOntology;
import org.semanticweb.owl.model.OWLOntologyCreationException;
import org.semanticweb.owl.model.OWLOntologyManager;
import org.semanticweb.owl.model.OWLSubClassAxiom;
import org.semanticweb.reasonerfactory.factpp.FaCTPlusPlusReasonerFactory;

//import uk.ac.manchester.cs.factplusplus.protege.FaCTPlusPlusReasonerFactory;


public class OWL_Query_Manager
{
    public static final String T_BOX_FILE = System.getenv( "T_BOX_FILE" );
    public static final String T_BOX_URI = System.getenv( "T_BOX_URI" );
    public static final String A_BOX_FILE= System.getenv( "A_BOX_FILE" );
    public static final String A_BOX_URI = System.getenv( "A_BOX_URI" );    
    // public static final String LOWER_ONTOLOGY_FILE= "/home/mccrae/ontologies/Ontology9999/Ontology9999.owl"; // System.getenv( "LOWER_ONTOLOGY_FILE" );
    // public static final String A_BOX_URI = "http://www.semanticweb.org/ontologies/Ontology9999.owl#"; // System.getenv( "A_BOX_URI" );
   
    OWLOntologyManager manager;
    OWLOntology t_box, a_box;
    OWLReasonerFactory factory; 
    OWLReasoner reasoner;    
    
    
    
    HashMap <OWLClass, Set <String>> lexicalisations;
    // Maps all concepts from the T-Box to the set of their lexicalisations.
    
    HashMap <String, Set <OWLClass>> conceptualisation_in_ontology;	
    // Maps all lexicalisations from the T-Box to the set of concepts that have this lexicalisation.
	
	HashMap <OWLIndividual, OWLClass> context_concept;
	// Maps a concept individual to the concept it instantiates.
	
	HashMap <OWLClass, Set<OWLIndividual>> context_individuals;
	// Maps a concept to its set of instantiations in the context model.
	
	HashMap <OWLClass, Set<String>> context_arguments; 
	// Maps any concept from the T-Box onto its set of argument roles.
	// Concepts which do not take any arguments map onto the empty string.

	HashMap <String, HashMap<String, String>> context_relation;		
	/* 
	 * Maps the string of the relation subject to a hashmap which takes the relation
	 * object names as key and returns the relation between relation subject and
	 * relation object in the format 'is_ROLE_for' as value. 
	 * Asserted 'has_ROLE' relations are converted to 'is_ROLE_for' relations.
	*/
	
	HashMap <OWLClass, OWLClass> concept_number;	
	// Maps every concept in the T-Box upon the ontology concept that represents its grammatical number.
	// Concepts which do not have a grammatical number defined as a superclass are mapped to the generic
	// concept 'Grammatical.Number'.

	HashMap <OWLClass, String> situation_arity;	
	// Maps any concept in the T-Box upon its situation arity (sem_val). 
	// Concepts which do not have an situation arity defined map onto the empty string.
	
	
    public OWL_Query_Manager()
   	{
        try 
        {
        	// Create an ontology manager in the usual way.
        	manager = OWLManager.createOWLOntologyManager();

            // Load copies of the T-Box and the A-Box each.

        	t_box = manager.loadOntology( URI.create( "file://" + T_BOX_FILE ) );
            
//        	System.err.println( "Loaded " + t_box.getURI() );

            a_box = manager.loadOntology( URI.create( "file://" + A_BOX_FILE ) );

//            System.err.println( "Loaded " + a_box.getURI() );

            
            // We need to create an instance of OWLReasoner. The OWLReasoner provides the basic
            // query functionality that we need, for example the ability obtain the subclasses
            // of a class etc.  The instance of the OWLReasoner can be obtained either via an 
            // instance of a FaCTPlusPlusReasonerFactory or a PelletReasonerFactory.
            
            integrate_Fact();
            
            // integrate_Pellet();
                        
            // We now need to load the ontology into the reasoner.  This is typically the
            // imports closure of an ontology that we're interested in.  In this case, we want
            // the imports closure of Upper.Ontology.owl.  
            // Note that no assumptions are made about the dependency of one ontology on 
            // another ontology.  This means that if we loaded just the pizza ontology 
            // (using a singleton set) then any imported ontologies would not automatically be loaded.

            // Obtain and load the imports closure of the pizza ontology:
 
//            Set<OWLOntology> T_BOX_imports_closure = manager.getImportsClosure( T_BOX );
//            reasoner.loadOntologies( T_BOX_imports_closure );
//            reasoner.classify();
            // Obtain and load the imports closure (i.e.: the set of all ontologies that need to 
            // be imported by the desired ontology):
                        
            Set<OWLOntology> imports_closure = manager.getImportsClosure( a_box );
            
//            System.err.println( "The import closure contained " + imports_closure.size() + " elements." );
                       
            reasoner.loadOntologies( imports_closure );
            
            reasoner.classify();
  
            //	System.out.println( "Classified the ontology." );
            
            // We can examine the expressivity of our ontology ( some reasoners do not support
            // the full expressivity of OWL):
           //  DLExpressivityChecker checker = new DLExpressivityChecker( importsClosure );
            // System.out.println( "Ontology's Expressivity: " + checker.getDescriptionLogicName() );

            // We can determine if the ontology is actually consistent.  
            // (If an ontology is inconsistent then owl:Thing is equivalent to owl:Nothing - i.e. 
            // there can't be any models of the ontology.)

//            boolean consistent = reasoner.isConsistent( t_box );
//            System.out.println( "Upper Ontology Consistent: " + consistent );
//            System.out.println( "\n" );
//            boolean consistent = reasoner.isConsistent( a_box );
//            System.out.println( "Lower Ontology Consistent: " + consistent );
         
            // get_Inconsistent_Classes();
            
            // print_Class_Set( get_Descendants( "Lexicalised.Concept" ) );
                        
            // get_Inferred_Individuals( "Situation" );

            // get_Inferred_Subclasses( "Ontological.Concept" );
            
            // get_Inferred_Types( "etw.malen_01" );
            
            // get_Subclass_Axioms( "Man" );
    
            extract_Ontology_Lexicalisations();
            
            set_Concept_Numbers();
            
            infer_Situation_Arities();

            // get_Lexicalisation( "Baker.f" );

            // get_Conceptualisation( "Lex_etw.malen" );

            // System.out.println( "Is compatible: " + is_Compatible( "Movie.Diva", "Woman") );

            extract_Individuals_in_Context();            
        }
        catch( UnsupportedOperationException exception ) 
        {
            System.out.println( "Unsupported reasoner operation." );
        }
        catch( OWLReasonerException ex ) 
        {
            System.out.println( "Reasoner error: " + ex.getMessage() );
        }
        catch ( OWLOntologyCreationException e ) 
        {
        	e.printStackTrace();
            System.out.println( "Could not load the ontology: " + e.getMessage() );
        }        
   	}
   	

//	private void integrate_Pellet()
//	{
//		// Pellet requires the Pellet libraries  (pellet.jar, aterm-java-x.x.jar) and the
//        // XSD libraries that are bundled with pellet: xsdlib.jar and relaxngDatatype.jar
//        // all of which need to be on CLASSPATH and eclipse's Java Buildpath.
//		
//		OWLReasonerFactory factory = new PelletReasonerFactory();
//		
//		System.out.println( "Got this far OK." );
//		
//		reasoner = factory.createReasoner( manager );
//	}

    
    
    /**
     * This method sets the concept number for every concept in the T-Box. Concept number
     * is determined based whether the class has a superclass 'Singular' or 'Plural' in the 
     * T-Box. If they don't, their concept number is set to 'bot'.
     */
    
	private void set_Concept_Numbers()
	{
		concept_number = new HashMap <OWLClass, OWLClass> ();		
		 	    
	    OWLClass singular_concept = get_OWL_Ontology_Manager().getOWLDataFactory().getOWLClass( URI.create( get_T_BOX_URI() + "Singular" ) );							
	    OWLClass plural_concept = get_OWL_Ontology_Manager().getOWLDataFactory().getOWLClass( URI.create( get_T_BOX_URI() + "Plural" ) );							
	    OWLClass grammatical_number_concept = get_OWL_Ontology_Manager().getOWLDataFactory().getOWLClass( URI.create( get_T_BOX_URI() + "Grammatical.Number" ) );							
		
		
		for ( OWLClass cls : t_box.getReferencedClasses() ) 	// Iterate over all classes in the T-Box.
		{
			Set <OWLDescription> set = cls.getSuperClasses( t_box );
			
			// System.err.println( "Class " + cls + " has direct superclasses " + set );
								
			if ( set.contains( singular_concept ) & ! set.contains( plural_concept ))
			{
				concept_number.put( cls,  singular_concept );

				// System.err.println( "Class " + cls + " has direct superclass 'Singular'." );
			}
			else if ( set.contains( plural_concept ) & ! set.contains( singular_concept ) )	
			{
				concept_number.put( cls,  plural_concept );
				// System.err.println( "Class " + cls + " has direct superclass 'Plural'." );
			}
			else
			{
				concept_number.put( cls, grammatical_number_concept );
				// System.err.println( "Class " + cls + " has no number indicating superclass." );
			}
		}
	}  
    
    /**
     * This method sets the situation arity for every concept in the T-Box. Situation arity
     * is determined based on the class a concept has been asserted in. Concretely,
     * this method is checking if a concept is a member of the descendant classes of the special helper 
     * classes in the T-Box which have been defined such as to contain only situation concepts of a 
     * particular situation arity. Presently, these classes are part of the asserted hierarchy. Ideally, they
     * should all be inferred to permit a purely semantic - rather than a syntactic - structuring of situation
     * concepts in the asserted hierarchy. Hence, if a concept is a member of a descendant class of the class, 
     * say, 'Takes.AGENT.THEME, then it is assigned the situation arity 'ag_th'. Note that the representation 
     * string for situation arity is intentionally fully analogous to that of a verb's semantic valence.
     */
    
	private void infer_Situation_Arities()
	{
		situation_arity = new HashMap <OWLClass, String> ();
		Set<Set<OWLClass>> set_of_sets = new HashSet <Set<OWLClass>> ();
		
		try
	    {
	       	set_of_sets = reasoner.getDescendantClasses( manager.getOWLDataFactory().getOWLClass( URI.create( T_BOX_URI + "Takes.AGENT" ) ) );
	       	Set<OWLClass> ag = OWLReasonerAdapter.flattenSetOfSets( set_of_sets );
	 			
	       	set_of_sets = reasoner.getDescendantClasses( manager.getOWLDataFactory().getOWLClass( URI.create( T_BOX_URI + "Takes.AGENT.RECIPIENT" ) ) );
	       	Set<OWLClass> ag_re = OWLReasonerAdapter.flattenSetOfSets( set_of_sets );

	       	set_of_sets = reasoner.getDescendantClasses( manager.getOWLDataFactory().getOWLClass( URI.create( T_BOX_URI + "Takes.AGENT.THEME" ) ) );
	       	Set<OWLClass> ag_th = OWLReasonerAdapter.flattenSetOfSets( set_of_sets );

	       	set_of_sets = reasoner.getDescendantClasses( manager.getOWLDataFactory().getOWLClass( URI.create( T_BOX_URI + "Takes.AGENT.RECIPIENT.THEME" ) ) );
	       	Set<OWLClass> ag_re_th = OWLReasonerAdapter.flattenSetOfSets( set_of_sets );

	       	set_of_sets = reasoner.getDescendantClasses( manager.getOWLDataFactory().getOWLClass( URI.create( T_BOX_URI + "Takes.AGENT.THEME.THEME" ) ) );
	       	Set<OWLClass> ag_th_th = OWLReasonerAdapter.flattenSetOfSets( set_of_sets );
		
			for ( OWLClass concept : t_box.getReferencedClasses() )	// Iterate over all concepts in the T-Box.
			{
				String arity = "";
							
				// System.out.println( "\nChecking concept " + concept + " for its situation arity." );
				
				if ( ag.contains( concept ) )	
				{
					arity = "ag";
				}
				else if ( ag_re.contains( concept ) )
				{
					arity = arity + "ag_re";
				}
				else if ( ag_th.contains( concept ) )
				{
					arity = arity + "ag_th";
				}
				else if ( ag_re_th.contains( concept ) )
				{
					arity = arity + "ag_re_th";
				}
				else if ( ag_th_th.contains( concept ) )
				{
					arity = arity + "ag_th_th";
				}
				
				set_Situation_Arity( concept, arity );
				
				// System.out.println( "Detected situation arity '" + arity + "' for concept '" + concept.toString() + "'." );
			}
	    }
        catch ( OWLReasonerException e )
        {
			e.printStackTrace();
            System.out.println( "Could not determine the descendant classes for the situation concepts because: \n" + e.getMessage() );
		}
	}


	private void set_Situation_Arity( OWLClass concept, String arity )
	{
		situation_arity.put( concept, arity );
	}

	
	public String get_Situation_Arity( OWLClass concept )
	{
		String result = situation_arity.get( concept );
		
		if ( result == null )
		{
			result = "";
		}
		
		return result;
	}
	

	private void integrate_Fact()
	{
		// FaCT++ requires the libraries ( uk.ac.manchester.cs.owl.factplusplus-v1.2.0.jar, 
		// FaCTpp-OWLAPI-v1.2.0.jar, FaCT++-v1.2.0/owlapi-bin.jar and 
		// org.protege.editor.owl.jar) all of which need to be on CLASSPATH and 
		// eclipse's Java Buildpath.

		factory = new FaCTPlusPlusReasonerFactory();		
	    reasoner = factory.createReasoner( manager );
	}


	private void get_Inferred_Types( String name )
	{
		OWLIndividual individual = manager.getOWLDataFactory().getOWLIndividual( URI.create( A_BOX_URI + name ) );
		Set<Set<OWLClass>> result = null;
		
		try
		{
			System.out.println( "Types for individual '" + name + "':" );
			// The boolean parameter determines whether only the asserted type (true) 
			// or all inferred types (false) are returned.
			result = reasoner.getTypes( individual, true );
			print_Set_of_Sets( result );
		}
		catch ( Exception e)
		{
			e.printStackTrace();
            System.out.println( "Could not determine the types for individual '" + name + "' because: \n" + e.getMessage() );
		}
	}


	private void get_Inconsistent_Classes()
	{
        // We can easily get a list of inconsistent classes.  ( A class is inconsistent if it
        // can't possibly have any instances ).  Note that the getInconsistentClasses method
        // is really just a convenience method for obtaining the classes that are equivalent
        // to owl:Nothing.
		
		try
		{
			Set<OWLClass> inconsistent_classes = reasoner.getInconsistentClasses();

			if ( ! inconsistent_classes.isEmpty() ) 
			{
				System.out.println( "The following classes are inconsistent: " );
				
				for( OWLClass owl_class : inconsistent_classes ) 
				{
					System.out.println( "\t" + owl_class );
				}
			}
			else 
			{
				System.out.println( "There are no inconsistent classes." );
			}
		}
		catch( OWLReasonerException e )
		{
        	e.printStackTrace();
            System.out.println( "Could not determine inconsistent classes because: \n" + e.getMessage() );
		}
	}
		
	
	private void get_Inferred_Subclasses( String class_name )
	{
        OWLClass owl_class = manager.getOWLDataFactory().getOWLClass( URI.create( T_BOX_URI + class_name ) );
        
        try
        {
        	Set< Set<OWLClass> > classes = reasoner.getSubClasses( owl_class );

       		System.out.println( "Subclasses of class '" + class_name + "': " );
       		print_Set_of_Sets( classes );        	
        }
        catch ( OWLReasonerException e )
        {
        	e.printStackTrace();
            System.out.println( "Could not determine the subclasses for class '" + class_name + "' because: \n" + e.getMessage() );    	
        }
	}


	private Set<OWLClass> get_Descendants( String class_name )
	{
		// First get a reference to the input class.
        OWLClass owl_class = manager.getOWLDataFactory().getOWLClass( URI.create( T_BOX_URI + class_name ) );
        Set<OWLClass> result = null;
        // Now use the reasoner to obtain the subclasses of this reference.  Note the reasoner
        // returns a set of sets.  Each set represents a subclass of the given class where the
        // classes in the set represent equivalence classes.  For example, if we asked for the
        // subclasses of A and got back {{B, C}, {D}} then A would have essentially two subclasses.
        // One of these subclasses would be equivalent to the class D, and the other would be the class that
        // was equivalent to class B and class C.
        try
        {
        	Set< Set<OWLClass> > set_of_sets = reasoner.getDescendantClasses( owl_class );
            // In this case, we don't particularly care about the equivalences, so we will flatten this
            // set of sets and print the result
        	
            System.out.println( "All descandants of class '" + class_name + "': ");
            result = OWLReasonerAdapter.flattenSetOfSets( set_of_sets );
        }
        catch ( OWLReasonerException e )
        {
			e.printStackTrace();
            System.out.println( "Could not determine the individuals for class '" + class_name + "' because: \n" + e.getMessage() );
		}
        
        return result;
	}


	private void print_Set_of_Sets( Set< Set<OWLClass> > set_of_sets )
	{
        Set<OWLClass> set = OWLReasonerAdapter.flattenSetOfSets( set_of_sets );

        if ( ! set.isEmpty() )
        {
        	for( OWLClass owl_class : set ) 
        	{
        		System.out.println( "\t" + owl_class );
        	}

        	System.out.println( "\n" );		
        }
        else
        {
        	System.out.println( "\tThe set of sets is empty." );
        }

		System.out.println( "\n" );		
	}


    /**
     * 	This method retrieves all inferred instances of a class whose name 
     *	is provided as the input string.
     */

	private Set <OWLIndividual> get_Inferred_Individuals( String class_name )
	{
		
		Set <OWLIndividual> result = null;
		
		try
		{
			OWLClass owl_class = manager.getOWLDataFactory().getOWLClass( URI.create( T_BOX_URI + class_name ) );
			
			result = reasoner.getIndividuals( owl_class, false );
			
			System.out.println( "Direct and indirect instances of class '" +  class_name + "': " );

			print_Set( result );
			
		}
		catch ( OWLReasonerException e )
		{
			e.printStackTrace();
            System.out.println( "Could not determine the individuals for class '" + class_name + "' because: \n" + e.getMessage() );
		}
		
		return result;
	}


	private void print_Set( Set <OWLIndividual> input_set )
	{
		if ( ! input_set.isEmpty() )
		{
			for( OWLIndividual element : input_set ) 
			{
				System.out.println( "\t" + element );
			}
		}
		else
		{
        	System.out.println( "\tThe set is empty." );			
		}
		
		System.out.println( "\n" );
	}
	
	
	private void print_Class_Set( Set <OWLClass> input_set )
	{
		if ( ! input_set.isEmpty() )
		{
			for( OWLClass element : input_set ) 
			{
				System.out.println( "\t" + element );
			}
		}
		else
		{
        	System.out.println( "\tThe set is empty." );			
		}
		
		System.out.println( "\n" );
	}

	
	private void print_SubClassAxiom_Set( Set <OWLSubClassAxiom> input_set )
	{
		if ( ! input_set.isEmpty() )
		{
			for( OWLSubClassAxiom element : input_set ) 
			{
				System.out.println( "\t" + element );
			}
		}
		else
		{
        	System.out.println( "\tThe set is empty." );			
		}
		
		System.out.println( "\n" );
	}
	

	private void print_IndividualAxiom_Set( Set <OWLIndividualAxiom> input_set )
	{
		if ( ! input_set.isEmpty() )
		{
			for( OWLIndividualAxiom element : input_set ) 
			{
				System.out.println( "\t" + element );
			}
		}
		else
		{
        	System.out.println( "\tThe set is empty." );			
		}
		
		System.out.println( "\n" );
	}

	
//	public void  set_Lexicalisation( String concept_name )
//	{
//		OWLClass lexicalised_concept = manager.getOWLDataFactory().getOWLClass( URI.create( T_BOX_URI + concept_name ) );
//			
//		Pattern pattern = Pattern.compile( "^SubClassOf\\(([\\wäöüßÄÖÜ.]+) ObjectHasValue\\(has_Lexicalisation ([\\wäöüßÄÖÜ]+)\\){2,2}$" );
//		
//		// Not sure if and how this regular expression could be simplified by using the locale "DE", "de".
//
//		Matcher matcher = null;
//		String cls = "";
//		String lex = "";
//			
//		// System.out.println( "\nSubclass Axioms for lexicalised concept '" + concept_name + "':" );
//		
//		for ( OWLSubClassAxiom axiom : T_BOX.getSubClassAxiomsForLHS( lexicalised_concept ) )
//		{
//			matcher = pattern.matcher( axiom.toString() );
//					
//			if ( matcher.find() )					
//			{
//				cls = matcher.group( 1 );
//				lex = matcher.group( 2 );
//				System.out.println( "Class:" + cls );				
//				System.out.println( "Lexicalisation:" + lex + "\n" );				
//				System.out.println( "Axiom: " + axiom.toString() );
//				lexicalisation.put( lexicalised_concept, lex );
//				// System.out.println( "Set context lexicalisation of '" + concept_name + "' to '" + lex + "'." );
//				return;
//			}
//		}
//	}

	
	/** 
	 * 	This method extracts the set of asserted lexicalisations for all concepts in the T-Box. 
	 * 	Lexicalisation is extracted from the axiom assertions of the descendant concepts in the 
	 * 	inferred class 'Lexicalised.Concept'. Lexicalisation and the corresponding concept are 
	 * 	saved as key-value pairs in the HashMaps 'lexicalisation' and 'conceptualisation', 
	 * 	respectively. Note that since we extract lexicalisation from asserted axioms for a certain
	 * 	subset of T-Box classes, no inferred properties such as inherited lexicalisations from
	 * 	superclasses will be found. The latter would greatly complicate the issue.
	 */
	
	private void  extract_Ontology_Lexicalisations()
	{
		lexicalisations = new HashMap <OWLClass, Set <String>> ();
    	conceptualisation_in_ontology = new HashMap <String, Set <OWLClass>> ();
    	
		/*	Note that the axiom patterns from which the lexicalisation is extracted differ depending on whether the axiom
		 *	refers to an Entity.Concept or a Situation.Concept. Also, some concepts may have multiple lexicalisations.
    	 *	The following regular expression matches all three axiom types.
    	 *
    	 *	Multiple lexicalisations:
    	 *	Axiom: SubClassOf(Bouquet ObjectUnionOf(ObjectHasValue(has_Lexicalisation Blumenstrauss) ObjectHasValue(has_Lexicalisation Strauss)))
		*/
    	
		Pattern pattern = Pattern.compile( "^SubClassOf\\(([\\w.]+) .*ObjectHasValue\\(has_Lexicalisation ([\\w]+)\\)+$" );


		Matcher matcher = null;

		String lex = "";
			
		OWLClass owl_class = manager.getOWLDataFactory().getOWLClass( URI.create( T_BOX_URI + "Lexicalised.Concept" ) );
		Set <Set<OWLClass>> set_of_sets = new HashSet<Set<OWLClass>> ();
		
		try
		{
			set_of_sets = reasoner.getDescendantClasses( owl_class );	// Get the descendant concepts for the class 'Lexicalised.Concept'.
		}
		catch ( OWLReasonerException e)
		{
        	e.printStackTrace();
            System.err.println( "Could not get the descendant classes of the concept 'Lexicalised.Concept' because: \n" + e.getMessage() );
		}
		
		Set <OWLClass> lexicalised_concepts = OWLReasonerAdapter.flattenSetOfSets( set_of_sets );
//		System.err.println( "There are " + lexicalised_concepts.size() + " lexicalised concepts in the T-Box." );
		
        for ( OWLClass concept : lexicalised_concepts )	// Now iterate over all lexicalised concepts from the T-Box.
		{			
//			System.err.println( "\nSubclass Axioms for lexicalised concept '" + concept.toString() + "':" );
			
			for ( OWLSubClassAxiom axiom : t_box.getSubClassAxiomsForLHS( concept ) )	// Iterate over all axioms for each concept.
			{
				/*
				 * 	Note that concepts of type 'Concept.Name.m.f' have not been lexicalised explicitly
				 * 	in the T-Box. These concepts still have one or more 'has_Lexicalisation' relations, 
				 * 	however, which they inherit in the concept hierarchy. Clearly, we want to exclude 
				 * 	these concepts since they have no genuine lexicalisation of their own.
				 * 	
				 * 	Since axioms for these concepts contain the substring "ObjectUnionOf" resulting from
				 * 	the logical OR in those concepts' lexicalisation definition we need to to exclude 
				 * 	these concepts in addition to matching the above regular expression.
				 * 
				 */

				matcher = pattern.matcher( axiom.toString() );

//				System.err.println( "Axiom: " + axiom.toString() );
				
				if ( matcher.find() & ! axiom.toString().contains( "ObjectUnionOf" ) )					
				{
					lex = matcher.group( 2 );
				
					add_Lexicalisation( concept, lex );
									
					add_Conceptualisation( lex, concept );					
				}
			}
		}
	}

	
	/**	This method adds the input concept to the set of conceptualisations of the 
	 * 	input string.
	 * 
	 * @param string	String for which a conceptualisation is being added.
	 * @param concept	Concept to add as conceptualisation.
	 */
		
	private void add_Conceptualisation( String string, OWLClass concept )
	{		
		if ( conceptualisation_in_ontology.get( string ) == null )
		{
			conceptualisation_in_ontology.put( string, new HashSet <OWLClass> () );
		}

		conceptualisation_in_ontology.get( string ).add( concept );

//		System.err.println( "The string '" + string + "' now has the set of conceptualisations " + conceptualisation_in_ontology.get( string ) + "." );
	}


	/**	This method adds the input string to the set of lexicalisations of the 
	 * 	input concept.
	 * 
	 * @param concept	Concept for which the lexicalisation is defined.
	 * @param string	String which is being set as lexicalisation.
	 */
	
	private void add_Lexicalisation( OWLClass concept, String string )
	{
		if ( ! lexicalisations.keySet().contains( concept ) )
		{
			lexicalisations.put( concept, new HashSet <String> () );
		}
		
		lexicalisations.get( concept ).add( string );
		
//		System.err.println( "Concept '" + concept + "' now has the set of lexicalisations " + get_Lexicalisations( concept ) + "." );

	}


	/**
	 * By regex-parsing subclass axioms for all lexicalised concepts, this class searches 
	 * for the class which has the input string as lexicalisation value.
	 * 
	 * @param lex is the string which needs to be the concept's lexicalisation.
	 * 
	 */
	
//	public void set_Context_Conceptualisation( String lex )
//	{
//		Pattern pattern = Pattern.compile( "^SubClassOf\\(([\\wäöüßÄÖÜ.]+) .*ObjectHasValue\\(has_Lexicalisation " + lex + "\\)+$" );
//		// Not sure if and how this regular expression could be simplified by using the locale "DE", "de".
//		
//		Matcher matcher = null;
//		String result = "";
//		
//		for ( OWLClass lexicalised_concept : get_Descendants( "Lexicalised.Concept" ) )	// Loop over all concepts which have been inferred to have a lexicalisation.
//		{
//			// System.out.println( "\nSubclass Axioms for lexicalised class '" + lexicalised_concept + "':" );
//		
//			for ( OWLSubClassAxiom axiom : T_BOX.getSubClassAxiomsForLHS( lexicalised_concept ) )	// Loop over all subclass axioms for a lexicalised concept.
//			{
//				// System.out.println( "Axiom: " + axiom.toString() );
//
//				matcher = pattern.matcher( axiom.toString() );
//					
//				if ( matcher.find() )	// Check if the axiom matches the above regular expression.
//				{
//					System.out.println( "Class:" + matcher.group( 1 ) );				
//					conceptualisation.put( lex, lexicalised_concept );
//					System.out.println( "Setting conceptualisation of '" + lex + "' to '" + lexicalised_concept.toString() + "'." );
//					return;
//				}
//			}
//		}			
//	}

	
	/** This method returns the set of concepts from the T-Box 
	 * 	whose set of lexicalisations contains the input string.
	 * 
	 */
	
	public Set <OWLClass> get_Conceptualisations_from_Ontology( String lex )
	{
		Set <OWLClass> result = new HashSet <OWLClass> ();
		
		if ( conceptualisation_in_ontology.keySet().contains( lex ) )
		{
			result = conceptualisation_in_ontology.get( lex );
		}

//		System.err.println( "The set of conceptualisations for '" + lex + "' is: " + conceptualisation_in_ontology.get( lex ) + "." );

		return result;
	}

	
//	public Set <OWLSubClassAxiom> get_Subclass_Axioms( String class_name )
//	{
//		OWLClass owl_class = manager.getOWLDataFactory().getOWLClass( URI.create( T_BOX_URI + class_name ) );
//
//		Set <OWLSubClassAxiom> result = T_BOX.getSubClassAxiomsForLHS( owl_class );
//
//		if ( ! result.isEmpty() )
//		{
//			Pattern pattern = Pattern.compile( "has_Lexicalisation (\\w+)\\)" ); 
//			Matcher matcher = null;
//	
//			// System.out.println( "\nSubclass Axioms:" );
//
//			for ( OWLSubClassAxiom axiom : result )
//			{
//				matcher = pattern.matcher( axiom.toString() );
//			
//				if ( matcher.find() )					
//				{
//					// System.out.println( "Lexicalisation:" + matcher.group( 1 ) );				
//				}
//			}
//		}
//		return result;
//	}

	
	/**
	 * This method extraxts the individuals asserted in the context model as well as 
	 * their classes and relations. Effectively, this method therefore sets the
	 * conceptualisations of the individuals in the context model. 
	 * 
	 * Ideally, this method would return the contents of an inferred class in the 
	 * ontology that contains all classes which have an instantiation in the A-Box. 
	 * At present, however, I don't know how to define such a class restriction in OWL.
	 * 
	 * As a work-around, this method parses the set of axioms relating to individuals in 
	 * the ontology and then determines from those axioms the set of all individuals as 
	 * well as their classes in the A-Box.
	 * 
	 */
	
	public void extract_Individuals_in_Context()
	{	
		
//		System.err.println( "The A-Box contains instances of the concepts: " + a_box.getClassAxioms() );
		context_concept = new HashMap <OWLIndividual, OWLClass> ();
		context_individuals = new HashMap <OWLClass, Set<OWLIndividual>> ();
		context_relation = new HashMap <String, HashMap<String, String>> ();
    	context_arguments = new HashMap <OWLClass, Set<String>> ();

    	
    	// Simple OWL class assertion: ClassAssertion(Etw.Versorgen Etw.Versorgen_01)
    	// Complex OWL Class assertion: ClassAssertion(ObjectIntersectionOf(Man Singular) Man_01)
    	
		Pattern class_pattern = Pattern.compile( "^ClassAssertion\\(([\\w. ]+) ([\\w.0-9]+)\\)$" );
		Matcher class_matcher = null;
		
		// Note that thematic relations may be specified in the context model via two relations - 
		// either 'is_ROLE_for' or via 'has_ROLE'.
		
		Pattern property_pattern = Pattern.compile( "^ObjectPropertyAssertion\\((has_[A-Z]+|is_[A-Z]+_for) ([\\w.0-9]+) ([\\w.0-9]+)\\)$" );
		Matcher property_matcher = null;
				
		for ( OWLIndividualAxiom axiom : a_box.getIndividualAxioms() )	// Iterate over all axioms relating to individuals in the ontology.
		{
//			 System.err.println( "A-Box OWLIndividualAxiom: " + axiom.toString() + " is of type " + axiom.getAxiomType() + " and has referents " + axiom.getReferencedEntities() + " and Signature " + axiom.getSignature() + ".\n" );
						
			class_matcher = class_pattern.matcher( axiom.toString() );
			property_matcher = property_pattern.matcher( axiom.toString() );
			
			/* 
			 * 	TO-DO: Include the correct handling of individuals for which no class assertion has been made in the 
			 * 	A-Box. These individuals should be classified as instantiations of OWL:Thing.
			*/	
			
			if ( class_matcher.find() ) // If the axiom is a class assertion, extract the individual and its class from the axiom.
			{				
//				System.err.println( "Found concept " + class_matcher.group( 1 ) + " instantiated in the A-Box.");
				
				OWLClass cls = manager.getOWLDataFactory().getOWLClass( URI.create( T_BOX_URI + class_matcher.group( 1 ) ) );	
				
				OWLIndividual individual = manager.getOWLDataFactory().getOWLIndividual( URI.create( A_BOX_URI + class_matcher.group( 2 ) ) );

				add_Context_Concept( individual, cls );
								
				add_Context_Individual( cls, individual );
				
//				System.err.println( "Context individual '" + individual.toString() + "' has been added to the set of context instantiations of concept '" + cls + "'." );	
			}			
			else if ( property_matcher.find() ) // If the axiom is an object property assertion, extract the relation and the subject and object indiduals between which the relation holds.
			{
				String property_name = property_matcher.group( 1 );
				String subject_name =  property_matcher.group( 2 ); 				
				String object_name =  property_matcher.group( 3 );
				String temp = "";
				
				/* 
				 * Note that the following manipulations are a bit of a hack. Ideally, all 'is_ROLE_for' relations
				 * should be extracted from the set of inferred individual axioms. At present, however, this information
				 * can obly be obtained from the Pellet reasoner which has not yet been integrated successfully.
				 * Strictly, the modelling knowledge that 'has_ROLE' is the inverse property of 'is_ROLE_for' should
				 * not have been hard coded here any may be removed at a later stage.
				 */

				// Since we are primarily interested in relations of type 'is_ROLE_for', 'has_ROLE'
				// relations need to be re-written as 'is_ROLE_for' relations. To this end, rename the
				// relation and reverse subject and object individuals extracted.
				
				if ( property_name.indexOf( "has_" ) == 0 )
				{
					property_name = "i" +  property_name.substring( 2 ) + "_for";

					temp = subject_name;
					subject_name = object_name;
					object_name = temp;
				}
				
				set_Context_Relation( subject_name, object_name, property_name );
				
//				System.err.println( "Context individual '" + subject_name + "' " + property_name + " context individual '" + object_name + "'." );
				
				Pattern situation_pattern = Pattern.compile( "AGENT|THEME|RECIPIENT" );
				Matcher situation_matcher = null;
				situation_matcher = situation_pattern.matcher( property_name );
				
				if ( situation_matcher.find() )		// Check if the object extracted is an instance of a situation concept.
				{
					if ( ! context_arguments.keySet().contains( object_name ) )	// This is the first relation extracted for this situation individual.
					{
						Set <String> set = new HashSet <String> ();
						context_arguments.put( manager.getOWLDataFactory().getOWLClass( URI.create( T_BOX_URI + object_name ) ), set );
					}

					context_arguments.get( manager.getOWLDataFactory().getOWLClass( URI.create( T_BOX_URI + object_name ) ) ).add( property_name );
					
					// System.out.println( "Added property '" + property_name + "' to set of relations for context individual '" + object_name + "'." );
				}
			}
		}
	}

	
	private void add_Context_Concept( OWLIndividual individual, OWLClass cls )
	{
		context_concept.put( individual, cls );		
	}


	private void add_Context_Individual( OWLClass cls, OWLIndividual individual )
	{
		// If no set of context individuals has been created for this class yet, create it.
		
		if ( context_individuals.get( cls ) == null )	
		{
			context_individuals.put( cls, new HashSet <OWLIndividual> () );
		}

		context_individuals.get( cls ).add( individual );		
	}

	

	/**
	 * This method returns the set of concepts that have an assserted 
	 * instantiation in the A-Box (the context model).
	 *  
	 */
	
	public Set <OWLClass> get_Context_Concepts()
	{
		// System.out.println( "There are " + context_concept.size() + " concepts in the context." );
		return context_individuals.keySet();
	}	

	
	/**
	 * This method returns the concept which in the A-Box 
	 * (the context model) is asserted to be instantiated 
	 * by the given context individual.
	*/
	
	public OWLClass get_Context_Concept( OWLIndividual individual )
	{
		return context_concept.get( individual );
	}
	
	
	
	/**
	 * This method returns the set of all individuals that have been 
	 * assserted in the A-Box (the context model).
	 *  
	 */
	
	public Set <OWLIndividual> get_All_Context_Individuals()
	{
		return OWLReasonerAdapter.flattenSetOfSets( (Set <Set<OWLIndividual>>) context_individuals.values() );
	}

	

	/**
	 * This method returns the set of context individuals that have been 
	 * assserted in the A-Box (the context model) for the given input 
	 * class.
	 */

	public Set<OWLIndividual> get_Context_Individuals( OWLClass input_class )
	{
		return context_individuals.get( input_class );
	}


	
	/**
	 * This method returns the grammatical number concept assigned to the 
	 * input concept.
	 *  
	 */
	
	public OWLClass get_Grammatical_Number( OWLClass cls )
	{
		return concept_number.get( cls );
	}	
	
	
	
//	/**
//	 * This method returns the set of context individuals that have been 
//	 * assserted in the A-Box (the context model) for the given word. 
//	 */
//
//	public Set <OWLIndividual> get_Context_Individuals( Word word )
//	{
//		HashSet <OWLIndividual> result = new HashSet <OWLIndividual> ();
//		
//		for ( OWLClass concept : word.get_Conceptualisations() )
//		{
//			result.addAll( context_individuals.get( concept ) );
//		}
//		
//		return result;
//	}	

	
	private void set_Context_Relation( String individual_name_1, String individual_name_2, String relation )
	{	
		if ( context_relation.get( individual_name_1 ) == null )
		{
			context_relation.put( individual_name_1, new HashMap<String, String> () );
		}

		if ( individual_name_1 != null && individual_name_2 != null && relation != null )
		{
			context_relation.get( individual_name_1 ).put( individual_name_2, relation );		
		}
		else
		{
			System.out.println( "Cannot generate the relation '" + relation + "' for individuals '" + individual_name_1 + "' and '" + individual_name_2 + "'." );
		}
	}


	/**
	 * This method returns the name of the contextually asserted relation 
	 * between two contextually asserted individuals. Note that in the current 
	 * implementation we assume that only one relation has been asserted for a 
	 * pair of individuals ... this, clearly, is a simplification. If no relation 
	 * has been asserted or one of the input individuals has not been asserted 
	 * in the context model, the method returns an empty string.
	 * 
	 */
	
	public String get_Context_Relation( OWLIndividual ind_1, OWLIndividual ind_2 )
	{
		// System.out.println( "Checking for context relation between dependant: " + ind_1 + " and regent: " + ind_2 + "." );
		
		String result = "";
		
		if ( ind_1 != null && ind_2 != null && context_relation.get( ind_1.toString() ) != null )
		{
			result = context_relation.get( ind_1.toString() ).get( ind_2.toString() );

			if ( result == null )
			{
				result = "";
			}
		}
		
		return result;
	}
	
	
	public Set <String> get_Lexicalisations( OWLClass owl_class )
	{		
		// System.err.println( "The set of lexicalisations for concept '" + owl_class.toString() + "' is: '" + lexicalisations.get( owl_class ) + "'." );
		return lexicalisations.get( owl_class );
	}

	
	public Set <String> get_Lexicalisations( String class_name )
	{		
		System.err.println( "The set of lexicalisations for concept '" + class_name + "' is: '" + lexicalisations.get( class_name ) + "'." );
		OWLClass owl_class = manager.getOWLDataFactory().getOWLClass( URI.create( T_BOX_URI + class_name ) );
		return lexicalisations.get( owl_class );
	}

	
	/** This method determines if an individual can be the member of both input classes
	 * 	without making the ontology inconsistent, i.e. if the properties asserted for
	 * 	each one of the input classes are compatible with each other. In description logic
	 * 	terms this means the reasoner checks if the intersection (AND conjunction) of the 
	 * 	two input classes is satisfiable in (i.e.: consistent with) the given T-Box.
	 * 
	 * @param class_1 the first input class
	 * @param class_2 the second input class
	 * @return boolean saying whether or not the two concept classes are compatible with eachother
	 */
	
	public boolean is_Compatible( OWLClass class_1, OWLClass class_2 )
	{
		boolean result = false;

		try
		{
			result = reasoner.isSatisfiable( manager.getOWLDataFactory().getOWLObjectIntersectionOf( class_1, class_2 ) );
			// System.err.println( "Compatibility of classes " + class_1 + " and class " + class_2 + ": " + result + ".");
		}
		catch ( OWLReasonerException e )
        {
        	e.printStackTrace();
            System.err.println( "Could not determine whether OWLClasses " + class_1 + " and " + class_2 + " are mutually compatible.\n" + e.getMessage() );
        }        
		
		return result;
	}
	
	
	public boolean is_Compatible( String class_name_1, String class_name_2 )
	{
        OWLClass class_1 = manager.getOWLDataFactory().getOWLClass( URI.create( T_BOX_URI + class_name_1 ) );
        OWLClass class_2 = manager.getOWLDataFactory().getOWLClass( URI.create( T_BOX_URI + class_name_2 ) );
		
		return is_Compatible( class_1, class_2 );
	}


	public OWLOntologyManager get_OWL_Ontology_Manager()
	{
		return manager;
	}


	public OWLOntology get_T_BOX()
	{
		return t_box;
	}	
	

	public String get_T_BOX_URI()
	{
		return T_BOX_URI;
	}	
}
