/*
 * Copyright (C) 1997-2010 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

/*
 * beamsearch.h
 *
 *  Created on: 02.08.2010
 *      Author: Niels Beuck
 */

#ifndef BEAMSEARCH_H_
#define BEAMSEARCH_H_

/* ---------------------------------------------------------------------------
 * @addtogroup Beamsearch
 * @{
 */

/* -- INCLUDES ------------------------------------------------------------- */
#include <config.h>
#include "constraintnet.h"
#include "transform.h"
#include "blah.h"
#include "parse.h"
#include "profiling.h"

/* -- MACROS --------------------------------------------------------------- */



/* -- TYPE DEFINITIONS ----------------------------------------------------- */

/* ----------------------------------------------------------------------------
 * A beam search item describes an entry in the beam of the beamsearch strategy
 * long description
 */
struct BeamSearchStateStruct {
  ConstraintNet net;					// shared between entries
  Vector		virtualNodeTypes; 		// contains pointers to virtual node specs
										//   index 0 in the BV corresponds to net->lg->ve->vStartId
  Vector		entries;			    // the beam
  ProfilingObject profiling;       // list of measure points for profiling
  struct BeamSearchStateStruct *previous; // link to a previous state
  String 		newWord; 				// word that was introduced most recently
};
typedef struct BeamSearchStateStruct BeamSearchStateStruct, *BeamSearchState;


/* ----------------------------------------------------------------------------
 * A beam search item describes an entry in the beam of the beam search strategy
 * long description
 */
struct BeamEntryStruct {
  Analysis 		best;					// analysis, corresponds to the parse
  Parse 		p;						// parse, corresponds to the analysis
  ByteVector	virtualNodeUsed;		// cached information about which virtual nodes are used,
										//   index to gn->no
  ByteVector	virtualNodeAvailable; 	// information about which virtual nodes from the shared constraint net is available
  struct BeamEntryStruct *previous;     // the beam entry this one was generated from, NULL for first step entries
  String        method;                 // how this beam entry was generated
  int           rank;		            // rank in the beam
  Vector		lexemeSelection;		// selected lexeme for every word, not as part of the output but as guide for derived beamentries
  Vector		slots;					// selected levelValues from the analysis
};
typedef struct BeamEntryStruct BeamEntryStruct, *BeamEntry;


struct TopDownPredictorStruct{
	String constraintName; // name of the constraint that acts as a predictor
	List specs;	      // list of vnspecs that could cure the conflict
	int edge;         // which edge participating in the constraint is related to the predictions
	Boolean bottom;   // whether the bottom word of the edge is related to the prediction
	List labels; 	  // the labels to be considered for attachment
};
// typedef TopDownPredictorStruct *TopDownPredictor;


/* -- VARIABLES ------------------------------------------------------------ */


/* -- FUNCTIONS ------------------------------------------------------------ */

extern BeamSearchState beamSearch(Lattice lat);
extern TopDownPredictor tdpNew(String name, int edge, Boolean bottom, List labels, List specs);
extern void bsInitialize();

/* ------------------------------------------------------------------------- */
/** @} */
#endif /* don't insert anything after this #endif */
