/*
 * Copyright (C) 1997-2011 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * profiling.c
 *
 *  Created on: 15.02.2011
 *      Author: beuck
 */

#include <string.h>
#include <blah.h>
#include "profiling.h"
#include "timer.h"

/*
 * allocates a new measurePoint
 * and sets name and time
 * no need to change it afterwards
 */
ProfilingMeasurePoint proMPointNew(String name, unsigned long time) {
	ProfilingMeasurePoint ret= (ProfilingMeasurePoint) memMalloc(sizeof (ProfilingMeasurePointStruct));
	ret->name=name;
	ret->time=time;
	return ret;
}

/*
 * constructor
 */
ProfilingObject proNew() {
	ProfilingObject pro = (ProfilingObject) memMalloc(sizeof (ProfilingObjectStruct));
	pro->measurePoints = NULL;
	pro->contraintCount = hashNew( 500, 0.8, hashStringHashFunction, hashStringEqualFunction );
	return pro;
}

/*
 * reads the elapsed time from the given timer (resets it)
 * and attaches a new measure point to the given list
 */
void proMeasureTime(Timer timer, ProfilingObject prof, String name) {
	ProfilingMeasurePoint mp;
	mp = proMPointNew(strRegister(name), timerElapsed(timer));
	prof->measurePoints = listAppendElement(prof->measurePoints, mp);
	timerStart(timer);
}

/*
 * adds times in the second profiling to times of the respective measure points in the first one
 */
void proMerge(ProfilingObject prof1, ProfilingObject prof2) {
	List l, m;
	ProfilingMeasurePoint mp1, mp2;
	Boolean found;

	// combine entries present in both list by adding up the measured time
	for(l= prof1->measurePoints; l; l=l->next) {
		mp1 = l->item;
		for(m= prof2->measurePoints; m; m=m->next) {
			mp2= m->item;
			if(strcmp(mp1->name, mp2->name) == 0) {
				mp1->time += mp2->time;
				break;
			}
		}
	}

	// add entries only present in the second list to the first
	for(l= prof2->measurePoints; l; l=l->next) {
		mp1 = l->item;
		found = FALSE;
		for(m= prof1->measurePoints; m; m=m->next) {
			mp2= m->item;
			if(strcmp(mp1->name, mp2->name) == 0) {
				found = TRUE;
				break;
			}
		}
		if(!found) {
			prof1->measurePoints= listAppendElement(prof1->measurePoints, mp1);
		}
	}

	// merge constraint count
	if(hashIsEmpty(prof2->contraintCount)) {
		// nothing to do
	} else {
		if(hashIsEmpty(prof1->contraintCount)) {
			prof1->contraintCount = prof2->contraintCount;
		} else {
			for( l= hashListOfKeys(prof2->contraintCount); l; l=l->next) {
				String s = l->item;
				if(hashContainsKey(prof1->contraintCount, s)) {
					hashSet( prof1->contraintCount, s,
								(void *) ((int)hashGet(prof1->contraintCount, s)) + ((int)hashGet(prof2->contraintCount, s)));
				} else {
					hashSet(prof1->contraintCount, s, hashGet(prof2->contraintCount, s));
				}
			}
		}
	}

}
