/* 
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
 
/* Functions tested in Hash Module 
 * hashNew
 * hashSet
 * hashGet
 * hashRemove
 * hashSize
 * hashIsEmpty
 * hashContainsKey
 * hashContainsValue
 * hashDelete
 * hashListOfKeys
 * hashIteratorNew
 * hashIteratorNextKey
 * hashIteratorNextValue
 * hashIteratorDelete
 * hashStringHashFunction
 * hashStringEqualFunction
 *
 * Not Tested
 *
 * hashForEach
 * hashForEachWithData
 * hashForEachFree
 */

#include <stdio.h>
#include <config.h>
#include <string.h>
#include <blah.h>
#include <assert.h>

//#define VERBOSE

int main(int argc, char **argv)
{
  Hashtable ht = NULL;
  String a[] =
    { "Mike", "Killian", "Ajay", "Sofiya", "Aarthi", "Krish", "Adarsh" };
  String outstring, valstr, keystr;
  int i;
  HashIterator hi;

  blahInitialize();

  /* --------------------------------------------- Test hashNew */

  ht = hashNew(20, 0.5, hashStringHashFunction, hashStringEqualFunction);

  for (i = 0; i < 4; i++) {
    (void)hashSet(ht, a[i], (Pointer) a[i]);
    assert(hashIsEmpty(ht) == FALSE);
  }

  /* --------------------------------------------- Testing hashGet */

  for (i = 0; i < 4; i++) {
    outstring = (String) hashGet(ht, a[i]);
    assert(outstring == a[i]);
  }

  /* --------------------------------------------- Testing hashSize */

  assert(hashSize(ht) == 4);

  for (i = 4; i < 7; i++) {
    (void)hashSet(ht, a[i], (Pointer) a[i]);
  }

  assert(hashSize(ht) == 7);

  /* --------------------------------------------- Testing haskContainsKey */

  for (i = 0; i < 7; i++) {
    assert(hashContainsKey(ht, a[i]) == TRUE);
  }

  /* --------------------------------------------- Testing HashContainsValue */

  {
    assert(hashContainsValue(ht, (Pointer) a[0]) == TRUE);
    assert(hashContainsValue(ht, (Pointer) a[2]) == TRUE);
    assert(hashContainsValue(ht, (Pointer) a[3]) == TRUE);
    assert(hashContainsValue(ht, (Pointer) a[5]) == TRUE);
  }

  /* --------------------------------------------- Testing hashListOfKeys */

  { 
    List keys = hashListOfKeys(ht);
    for (i = 0; i < 7; i++) {
      assert(listContains(keys, a[i]));
    }
    keys = NULL;
  }

  /* --------------------------------------------- Testing hashIterator */

  hi = hashIteratorNew(ht);

  valstr = hashIteratorNextValue(hi);
  assert(valstr == a[5]);

  valstr = hashIteratorNextValue(hi);
  assert(valstr == a[3]);

  valstr = hashIteratorNextValue(hi);
  assert(valstr == a[0]);

  valstr = hashIteratorNextValue(hi);
  assert(valstr == a[2]);

  keystr = hashIteratorNextKey(hi);
  assert(keystr == a[6]);

  keystr = hashIteratorNextKey(hi);
  assert(keystr == a[4]);

  keystr = hashIteratorNextKey(hi);
  assert(keystr == a[1]);

  /* --------------------------------------------- hashIteratorDelete */

  hi = NULL;

  /* --------------------------------------------- Testing hashRemove */

  {
    List keys;
    List l;
    
    assert(hashSize(ht) == 7);

    outstring = (String) hashRemove(ht, a[6]);
    assert(outstring == "Adarsh");
    assert(hashSize(ht) == 6);

    outstring = (String) hashRemove(ht, a[6]);
    assert(outstring == 0);
    assert(hashSize(ht) == 6);

    outstring = (String) hashRemove(ht, a[5]);
    assert(outstring == "Krish");
    assert(hashSize(ht) == 5);

    keys = hashListOfKeys(ht);
    assert(listSize(keys) == 5);
    for (l = keys; l; l = l->next) {
#ifdef VERBOSE
    fprintf(stderr, "DEBUG: key='%s' %p\n", l->item, l->item);
#endif
    }

    keys = NULL;
  }

  /* --------------------------------------------- hashDelete */
  ht = NULL;

  blahFinalize();

  return 0;
}
