/*
 * Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. For
 * more details read COPYING in the root of this distribution.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

/* this will print information about an existing berkeley database */

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include "blah.h"
#include "db.h"

static DB *database;


/* ------------------------------------------------------------------------- 
 * open a database.
 * this function opens a database, specified by the parameter 'filename'
 * and stores the handle in the static variable 'database'. 
 * @param filename the filename of the database file
 * @returns TRUE if success, FALSE otherwise
 */
Boolean dbOpen(String filename)
{
  int ret = 0;		     // return code for database operations

  // create database handle
  if ((ret = db_create(&database, NULL, 0)) != 0) {
    printf("ERROR (db_create): %s\n", db_strerror(ret));
    return FALSE;
  }
  // open database 'filename'
#if (DB_VERSION_MAJOR < 4 || \
     DB_VERSION_MAJOR == 4 && DB_VERSION_MINOR == 0)
  if ((ret = database->open(database, filename, NULL, DB_UNKNOWN, 0, 0664)) != 0) {
#else
  if ((ret = database->open(database, NULL, filename, NULL, DB_UNKNOWN, 0, 0664)) != 0) {
#endif

    printf("ERROR (database->open): %s\n", db_strerror(ret));
    return FALSE;
  }
//  printf("file succesfully opened\n");
  return TRUE;
}


/* ------------------------------------------------------------------------- 
 *
 * prints all entries from the database
 * @returns a list of all lexical Entries.
 * @returnsa NULL if there was an ERROR
 */

void dbPrintAll(void)
{
  int ret = 0;		     // return code for database operations
  DBT key, value;	     // DBT structs for key and value
  DBC *cursorp;		     // cursor for getting duplicates
  int pos[2];
  char keyString[200];
  
  if (!database) {
    printf("ERROR: No database opened\n");
    return;
  }
  // create cursor
  database->cursor(database, NULL, &cursorp, 0);
  memset(&key, 0, sizeof (key));
  memset(&value, 0, sizeof (value));

  // now get all entries
  while ((ret =
	  cursorp->c_get(cursorp, &key, &value,
			 DB_NEXT)) != DB_NOTFOUND) {
    pos[0] = *((int *)value.data);
    pos[1] = *((int *)value.data + 1);
    // add final \0 to key
    memcpy(keyString, key.data, key.size);
    keyString[key.size] = '\0';
    printf("%-10s --> [%d] [%d]\n", keyString, pos[0], pos[1]);
    if (ret != 0) {
      printf("ERROR (cursorp->c_get): %s\n", db_strerror(ret));
      break;
    }
  }
}


/* ------------------------------------------------------------------------- 
 * close the database
 * write database to disk and close it.
 * @returns TRUE if success, FALSE otherwise
 */

Boolean dbClose(void)
{
  int ret = 0;		     // return code for database operations

  if (database == NULL) {
    printf("ERROR: No open database found for closing\n");
    return FALSE;
  }

  if ((ret = database->close(database, 0)) != 0) {
    printf("ERROR (database->close): %s\n", db_strerror(ret));
    return FALSE;
  }
  return TRUE;
}
  

int main(int argc, String argv[]) {
  if(argc > 2 || argc == 1) {
    printf("wrong parameter count: please specify one single database file\n");
  }
  
  if (!dbOpen(argv[1]))
    exit(1);
  dbPrintAll();
  dbClose();
  return 0;
}
  


  
