#!/usr/bin/env perl
# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# cdg/utils/ext2db.pl
#
# Transform a file of cooccurrence frequencies into a file
# suitable for quick access via DB_File.
use warnings;
use strict;
use DB_File;

$| = 1;

my $filename = shift;
(my $dbname = $filename) =~ s/ext$/pp/;
if($dbname eq $filename) {
  die "Nice try! You won't get me to overwrite you valuable data *that* easily.\n";
}

our %h;
unlink $dbname;
tie %h, "DB_File", $dbname
  or die "Can't write `$dbname': $!\n";

open(EXT, "$filename") or die "Can't read `$filename': $!\n";
while(<EXT>) {
  if (/^sentences (\d+)/) {
    $h{sentences} = $1;
  } elsif (/words (\d+)/) {
    $h{words} = $1;
  } elsif (/^word +(\S+) +(\d+)/) {
    $h{"word|$1"} = $2;
  } elsif (/^tag (\S+) (\d+)$/) {
    $h{"tag|$1"} = $2
  } elsif (/^pp (\S+) (\S+) (\d+)$/) {
    $h{"pp|$1|$2"} = $3
  } elsif (/^pair (\S+) (\S+) (\d+)$/) {
    $h{"wpf|$1|$2"} = $3;
  } elsif (/^tpair (\S+) (\S+) (\d+)$/) {
    $h{"tpf|$1|$2"} = $3;
  } elsif (/^triple (\S+) (\S+) (\S+) (\d+)$/) {
    $h{"wpp|$1|$2"} = $3;
  } elsif (/^ttriple (\S+) (\S+) (\S+) (\d+)$/) {
    $h{"tpp|$1|$2"} = $3;
  } elsif(/^ *#/) {
    # do nothing
  } else {
    warn "Can't deal with `$_'\n";
  }
  print "[$.]\r" if 0 == ($. % 10000);
}
close(EXT);
untie %h;


