#!/usr/bin/env perl
#
# Copyright (c) 2007 Ronald Landheer-Cieslak <ronald@landheer-cieslak.com>
# Copyright (c) 2008 Robin Sonefors <ozamosi@flukkost.nu>

use strict;
use utf8;
use Encode;

sub fetch_log {
        my $command='git log --pretty=format:"%ai|%an|%h|%s" ';
        foreach (@_) {
                $command .= ' ' . $_;
        }
        $command .= " |";
        my $fh;
        open $fh, $command
                or die "Failed to fetch logs";
        # logs are presented as lines in which fields are separated by pipes.
        # the fields are the date in ISO format (so the date as we want it extends to the
        # first space), the name of the author, the abbreviated SHA1 hash and the subject line
        my $date_time;
        my $date;
        my $cruft;
        my $author;
        my $hash;
        my $subject;
        my $prev_date="";
        my @cache; # a cache of the changes for the current date (i.e. while $prev_date eq $date)
        my %entry; # a cache entry; $entry{'author'} is the entry and @$entry{'changes'} are the changes for the author in question
        while (<$fh>) {
                ($date_time, $author, $hash, $subject) = split(/\|/);
                ($date, $cruft) = split(/\s/, $date_time, 2);
                chomp $author;

                if ($date ne $prev_date)
                {
                        foreach (@cache)
                        {
                                # print out the line with the date and the author
                                my $changes = $_->{'changes'};
                                print "\n" . $prev_date . "\t" . $_->{'author'} . "\n" if ($#{$changes} != -1);
                                foreach (@{$changes})
                                {
                                        if (!utf8::is_utf8($_->{'subject'}))
                                        {
                                                Encode::from_to($_->{'subject'}, "iso-8859-1", "utf-8");
	                                    }
                                        print "\t" . $_->{'hash'} . ': ' . $_->{'subject'};
                                }
                        }
                        $prev_date = $date;
                        @cache = ();
                }
                # try to find an entry with the same author in the cache
                my $found = -1;
                my $i;
                for ($i = 0; $i <= $#cache && $found == -1; $i++)
                {
                        if ($cache[$i]->{'author'} eq $author)
                        {
                                $found = $i;
                        }
                }
                if ($found == -1)
                {
                        my $changes = ();
                        push @cache, { 'author', $author, 'changes', $changes };
                        $found = $#cache;
                }
                push @{$cache[$found]->{'changes'}}, { 'hash', $hash, 'subject', $subject };
        }
}

fetch_log @ARGV;

