#!/usr/bin/env perl
# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# cdg/utils/make-extra-lexicon.pl
#
# Write an extra lexicon for lookup for TNT; e.g. the standard
# model of German shipped with it does not even recognize `seid'.
use warnings;
use strict;
BEGIN {
  use File::Basename;
  unshift @INC, dirname(readlink($0) || $0); 
  require 'cdgsetup.pl';
}
use CDG::Utils qw(:functions);

my $dir = $0;
$dir =~ s!/util.*$!/grammar/deutsch!;

chdir $dir
  or die "Can't find deutsch grammar! It is supposed to be in $dir.\n";

sub find_words {
  my $file = shift;

  open(LEX, $file) or die "Can't read `$file': $!\n";
  while(<LEX>) {
    if(my ($word,$cat) = /^(.+) *:= *\[.*\bcat *: *(\$[,\.\(]|[A-Z]+)/) {
      $word =~ s/^\s+|\s+$//;
      $cat =~ s/^\s+|\s+$//;
      $word =~ s/ /\001/g;
      $word = parse_string $word;
      $cat = parse_string $cat;
      print "$word\t1\t$cat\t1\n";
    }
  }
  close(LEX);
}

find_words 'Lexikon.cdg';
find_words 'deutsch-lexikon.cdg';
