# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
eval '(exit $?0)' && eval 'exec perl -S $0 ${1+"$@"}'
 & eval 'exec perl -S $0 $argv:q'
 if 0;

#
# Ingo Schrder
#

if ( $0 =~ m/.*\/([^\/]*)/ ) {
    $cmd=$1;
} else {
    $cmd=$0;
}

@defaulthosts = ( "nats3", "nats9" );
$ssh="/opt/bin/ssh";
$ping="/usr/sbin/ping";
$dir_counter=0;

$id=getlogin() eq "" ? $< : getlogin();

require "getopts.pl";
&Getopts('dg:hj:kl:s:');
@hosts = defined($opt_l) ? &ReadFromFile($opt_l) : @defaulthosts;
@jobs = defined($opt_j) ? &ReadFromFile($opt_j) : @ARGV;
$debug = defined($opt_d) ? 1 : 0;
$sleep = defined($opt_s) ? $opt_s : 5;

$no_of_finished_jobs=0;
$pwd=`pwd`;
chop($pwd);
printf "Current directory is \"%s\"\n", $pwd;

die "Usage: $cmd [-h] [-j jobfile] [-k] [-l hostfile] [-s sleeptime] [script...]\n"
  if defined($opt_h);

for ($i=0; $i<=$#hosts; $i++) {
  if (&PingHost($hosts[$i])) {
    $host_status[$i]="free";
    printf "Using host\#%d \"%s\"\n", $i, $hosts[$i];
  } else {
    printf "Ignoring dead host \"%s\"\n", $hosts[$i];
    $host_status[$i]="ignored";
  }
}

for ($i=0; $i<=$#jobs; $i++) {
  $job_status[$i] = "new";
  printf "Accepted job\#%d \"%s\"\n", $i, $jobs[$i];
}

if (defined($opt_k)) {
  &KillUser($id);
  exit 0;
}

#if (defined($opt_g)) {
#  &KillProcessGroups("nats9", $opt_g);
#  exit 0;
#}

&ScheduleLoop;

exit 0;

# ------------------------------------------------------------
sub Die {
  local($m)=@_;

  printf STDERR "%s: %s\n", $cmd, $m;
  exit 0;
}

# ------------------------------------------------------------
sub PingHost {
  local($h)=@_;

  return !system("$ping $h 2 >/dev/null 2>&1");
}

# ------------------------------------------------------------
sub Ssh {
  local($h, $c)=@_;

  if (&PingHost($h)) {
#    printf("Starting \"%s\" on %s...", $c, $h);
    return system("$ssh -x -o \"ConnectionAttempts 1\" -o \"FallBackToRsh no\" $h \"$c\"");
#    printf("\n");
  } else {
    printf "ERROR: Host %s seems to be dead.\n", $h;
    return 1;
  }
}

# ------------------------------------------------------------
sub ReadFromFile {
  local($n)=@_;
  local(@l);

  open(H, "<$n") || &Die("can't read from file \"$n\"");
  while (<H>) { chop($_); push @l, $_; }
  close(H);
  return @l;
}

# ------------------------------------------------------------
sub KillUser {
  local($u)=@_;

  foreach $h (@hosts) {
    printf "Next host is %s... ", $h;
    # system("$ssh $h pgrep -U $id");
    system("$ssh $h pkill -U $id") && printf "FAILED";
    printf "\n";
  }
}

# ------------------------------------------------------------
sub KillProcessGroups {
  local($host, $pid)=@_;
  local($h, $hs);

  printf "KillProcessGroups %s %s\n", $host, $pid;
  $hs=`$ssh $host pgrep -P $pid`;
  chop($hs);
  system("$ssh $host kill $pid");
 HOST:
  foreach $h (split(/\s+/, $hs)) {
    next HOST unless $h=~m/^\d+$/;
    &KillProcessGroups($host, $h);
  }
}

# ------------------------------------------------------------
sub StartJobOnHost {
  local($d, $j, $h)=@_;
  local($ret);

  printf "Starting \"%s\" in %s on %s...", $j, $d, $h;
  $ret=&Ssh($h, "cd $d; $j > STDOUT 2> STDERR < /dev/null & echo \\\$!>PID");
  printf "\n";
  return !$ret;
}

# ------------------------------------------------------------
sub ScheduleLoop {

  while ($no_of_finished_jobs<$#jobs+1) {
  HOSTS:
    for ($j=0; $j<=$#hosts; $j++) {
      $h=$hosts[$j];
      $hs=$host_status[$j];
      printf "Checking host\#%d \"%s\" -> %s...\n", $j, $h, $hs;
      next HOSTS unless $hs eq "free";
    START:
      for ($i=0; $i<=$#jobs; $i++) {
	$js=$job_status[$i];
	printf "Checking job\#%d -> %s...\n", $i, $js;
	next START unless $js eq "new";
	do {
	  $dir=sprintf("d%05d", $dir_counter++);
	} while (-e $dir);
	mkdir($dir, 0700);
	open(F, ">$dir/HOST") || &Die("can't write HOST file: $!");
	printf F "%s\n", $h;
	close(F);
	if (&StartJobOnHost("$pwd/$dir", $jobs[$i], $h)) {
	  $job_status[$i]=$dir;
	  $host_status[$j]=$dir;
	  $host_for_job[$i]=$j;
	} else {
	  printf "Ignoring host\#%d \"%s\" from now on!\n", $j, $hosts[$j];
	  $host_status[$j]="dead";
	}
	next HOSTS;
      }
    }
    printf "Sleeping %d seconds...\n", $sleep;
    sleep $sleep;
  JOBS:
    for ($i=0; $i<=$#jobs; $i++) {
      $js=$job_status[$i];
      next JOBS unless $js=~m/^d\d+$/;
      if (-e "$js/FINISHED") {
	printf "Finished job\#%d \"%s\" in %s\n", $i, $jobs[$i], $js;
	$job_status[$i]="finished";
	$host_status[$host_for_job[$i]]="free";
	$no_of_finished_jobs++;
      }
    }
  }
}

# ------------------------------------------------------------
# ------------------------------------------------------------
