# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## AllHierarchies - manage all hierarchies.
##
## \todo Variable and function naming is a complete mess here. Til now I
## just fixed the declarations to match the implementations. Furtheron all
## underscore/capitalized namings should be unified. No interface variable
## should only be named \c id. When it is a hierarchy's id call it for example
## \c hierId. 
## \todo the TclArray called \c hier is cononfusing me with the TclArray \c hierarchy.
## I know they contain different stuff but again: the naming is extraordinary bad.
## \todo the TclArray \c hier is overloaded: its information should be split into
## several arrays named by the third argument of the hier array.
## \todo the methods C_to_Tcllist() and ldelete() are not related to AllHierarchies.
## \todo private varialbes should be named starting with and underscore
##
## \author Dietmar Fnning, Kilian A. Foth (see also AUTHORS and THANKS for more)
## ----------------------------------------------------------------------------
class AllHierarchies {
  inherit ::itk::Widget

  
  # public methods
  public method init_data {}
  public method getAllHierarchyIds {};

  constructor {args} {}; ## \type TclList

  # private methods
  private method C_to_Tcllist {list}; ## \type List
  private method _init_data {}
  private method balanceNode {id type_id}; ## \type TclString, TclString
  private method center {id level}; ## \type TclString, TclNumber
  private method compute_fathers_and_sons {hierarchy type_id}; ## \type Hierarchy, TclString
  private method compute_graph {id}; ## \type TclString
  private method compute_hierarchy {hierarchy}; ## \type Hierarchy
  private method compute_positions {id}; ## \type TclString
  private method display_graph {id level nodes_current_lvl x_start y_start counter}; ## \type TclString, TclNumber, TclList, TclNumber, TclNumber, TclNumber
  private method drawHierarchy {id}; ## \type TclString
  private method drawSort {id type_id}; ## \type TclString, TclString
  private method labelEnter {w pointerx pointery}; ## \type TclCommand, TclNumber, TclNumber
  private method labelLeave {w}; ## \type TclCommand
  private method importNode {id type_id}; ## \type TclString, TclString
  private method initial_x_pos {id level}; ## \type TclString, TclNumber
  private method initialize_depth {id type_id}; ## \type TclString, TclString
  private method is_hierarchy_a_tree {hierarchy type_id}; ## \type Hierarchy, TclString
  private method ldelete {list type_id} ; ## \type TclList, TclString
  private method leaves_right {id String}; ## \type TclString, TclString
  private method max_type_depth {id type_id depth}; ## \type TclString, TclString, TclNumber
  private method new_order {id String}; ## \type TclString, TclString
  private method number_of_fathers_and_sons {hierarchy type_id}; ## \type Hierarchy, TclString
  private method position_leaves {id level}; ## \type TclString, TclNumber
  private method position_leaves2 {id type_id}; ## \type TclString, TclString
  private method position_of_type {hierarchy type_id}; ## \type Hierarchy, TclString
  private method set_level_array {id type_id}; ## \type TclString, TclString
  private method stringlength {id type_id}; ## \type TclString, TclString
  private method xpixels_per_character {id}; ## \type TclString

  # private variables
  private variable hierarchies; ## \type TclArray
  private variable maxlength 0
  private variable maxdepth 0
  private variable widths; ## \type TclArray
  private variable levels; ## \type TclArray
  private variable hier; ## \type TclArray
  private variable unit 10

  private variable canvas ""; ## \type TclCommand
  private variable recurse_necessary 0
  private variable current_y 1
  private variable current_x 1
  private variable max_y 1
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body AllHierarchies::constructor {args} {

  itk_component add frame {
    frame $itk_interior.frame \
	-borderwidth 2 \
	-relief flat
  } { keep -cursor }

  # tabnotebook of canvases
  itk_component add tabno {
    iwidgets::tabnotebook $itk_component(frame).tabno \
	-borderwidth 0
  } {
    keep -background -cursor
    keep -angle -bevelamount -equaltabs -raiseselect -tabbackground
    keep -backdrop -tabpos
  }

  # packing
  pack $itk_component(frame) -fill both -expand 1
  pack $itk_component(tabno) -fill both -expand 1 -side top

  # get the initial data
  _init_data
  
  eval itk_initialize $args
}

## ----------------------------------------------------------------------------
## call _init_data if necessary.
## ----------------------------------------------------------------------------
body AllHierarchies::init_data {} {
  global inputCurrentGrammar

  if {[InputStruct_xcdghierarchiesToDate_get $inputCurrentGrammar]} {
     return
  }

  .cdgmain busy compute "Init Hierarchies" [code $this _init_data]

}

## ----------------------------------------------------------------------------
## get data from the cdg tool.
## ----------------------------------------------------------------------------
body AllHierarchies::_init_data {} {
  global inputCurrentGrammar

  set tabno $itk_component(tabno)

  # empty notebook
  catch {$tabno delete 0 end}
  foreach id [array names hierarchies] {
    unset hierarchies($id)
  }

  # import data to Tcl
  set hash [InputStruct_hierarchies_get $inputCurrentGrammar]
  set m [hashListOfKeys $hash]
  for {set l $m} { $l != "NULL" } { set l [ListStruct_next_get $l ] } {
    set id [ListStruct_item_get $l]
    set h [hashGet $hash $id]
    set id [HierarchyStruct_id_get $h]
    set hierarchies($id) $h
    compute_hierarchy $h
  }

  # allocate one page per hierarchy
  foreach id [array names hierarchies] {
    set h $hierarchies($id)
    set page [$tabno add -label $id]
    itk_component add $id { 
      iwidgets::scrolledcanvas $page.cs 
   } {}
    pack $itk_component($id) -fill both -side top -expand 1
    drawHierarchy $id

     # fast scanning
     set canvas [$itk_component($id) component canvas]
     bind $canvas <ButtonPress-2> [code $canvas scan mark %x %y] 
     bind $canvas <B2-Motion> [code $canvas scan dragto %x %y] 
  }

  # select first page
  if {[array names hierarchies] == "" } { 
    # add an empty tab
    set page [$tabno add -label ""]
    set dummy [iwidgets::scrolledcanvas $page.cs]
    pack $dummy -fill both -side top -expand 1
  }
  $tabno select 0
  InputStruct_xcdghierarchiesToDate_set $inputCurrentGrammar 1
}

## ----------------------------------------------------------------------------
## gather all relevant data for displaying a hierarchy.
## ----------------------------------------------------------------------------
body AllHierarchies::compute_hierarchy {hierarchy} {

  compute_fathers_and_sons $hierarchy top
  number_of_fathers_and_sons $hierarchy top
  set id [HierarchyStruct_id_get $hierarchy]
  set hier($id,is_a_tree) 1
  is_hierarchy_a_tree $hierarchy top
  initialize_depth $id top
  max_type_depth $id top 0
  stringlength $id top
}

## ----------------------------------------------------------------------------
## stores types stringlength.
## ----------------------------------------------------------------------------
body AllHierarchies::stringlength {id type_id} {

  if { $type_id == "" } {
    set hier($id,$type_id,stringlength) 0.2
  } else {
    set hier($id,$type_id,stringlength) [string length $type_id]
  }

  # recurse:
  foreach value $hier($id,$type_id,sons) {
    stringlength $id $value
  }
}

## ----------------------------------------------------------------------------
## initialize the depth information.
## This method recurses over all types in a hierarchy and sets their
## depth information to zero.
## ----------------------------------------------------------------------------
body AllHierarchies::initialize_depth {id type_id} {

  set hier($id,$type_id,max_type_depth) 0

  # recurse
  foreach value $hier($id,$type_id,sons) {
    initialize_depth $id $value
  }
}

## ----------------------------------------------------------------------------
## compute maximal distance between type and 'top'.
## This function finds a type's depth in trees, and the length
## of the longest path from top to the type for non-trees:
## ----------------------------------------------------------------------------
body AllHierarchies::max_type_depth {id type_id depth} {

  if {$depth > $hier($id,$type_id,max_type_depth)} {
    set hier($id,$type_id,max_type_depth) $depth
  }
  incr depth

  # recurse
  foreach value $hier($id,$type_id,sons) {
    max_type_depth $id $value $depth
  }
}

## ----------------------------------------------------------------------------
## compute tree property.
## This method computes whether a hierarchy is a tree.
## This function is traversing recursivly the complete
## hierarchy in the following way:
## - if the number of fathers is greater than one then the hierarchy 
##   is not a tree.
## \param hierarchy a pointer to a HierarchyStruct
## \param type_id an entry in the hierarchy
## ----------------------------------------------------------------------------
body AllHierarchies::is_hierarchy_a_tree {hierarchy type_id} {

  set id [HierarchyStruct_id_get $hierarchy]
  if {$hier($id,$type_id,number_fathers) > 1} {
    set hier($id,is_a_tree) 0
    return 0
  }

  # recurse
  foreach value $hier($id,$type_id,sons) {
    is_hierarchy_a_tree $hierarchy $value
  }
}

## ----------------------------------------------------------------------------
## stores the number of successors and predecessors each type has
## ----------------------------------------------------------------------------
body AllHierarchies::number_of_fathers_and_sons {hierarchy type_id} {

  set id [HierarchyStruct_id_get $hierarchy]
  set hier($id,$type_id,number_sons) [llength $hier($id,$type_id,sons)]
  set hier($id,$type_id,number_fathers) [llength $hier($id,$type_id,fathers)]

  # recurse
  foreach value $hier($id,$type_id,sons) {
    number_of_fathers_and_sons $hierarchy $value
  }
}  

## ----------------------------------------------------------------------------
## Converts a C-list of types into a Tcl-list of type-ids and filters out 'bot'.
## Only this function and position_of_type work with C-data structures.
## ----------------------------------------------------------------------------
body AllHierarchies::C_to_Tcllist {list} {

  set result ""
  while {$list != "NULL"} {
    set item [pointer2string [ListStruct_item_get $list]]
    if {[string compare $item bot] != 0} {
      lappend result $item
    }
    set list [ListStruct_next_get $list]
  }
  return $result
}
      
## ----------------------------------------------------------------------------
## This is the interface to functions.i
## \todo the documentation does not match the intension of this function
## ----------------------------------------------------------------------------
body AllHierarchies::position_of_type {hierarchy type_id} {

  set list_of_types [hashListOfKeys [HierarchyStruct_types_get $hierarchy]]
  set type_coord 0
  while {$list_of_types != "NULL"} {
    set type_string [pointer2string [ListStruct_item_get $list_of_types]]
    if {[string compare $type_string $type_id] == 0} {
      break
    }
    incr type_coord 
    set list_of_types [ListStruct_next_get $list_of_types]
  }
  if {$list_of_types == "NULL"} {
    puts "function position_of_type: type '$type_id not' found!"
  }
  return $type_coord
}

## ----------------------------------------------------------------------------
## find a types sons and fathers in the hierarchy.
## (uses the C-function listOfSons)
## ----------------------------------------------------------------------------
body AllHierarchies::compute_fathers_and_sons {hierarchy type_id} {

  set id [HierarchyStruct_id_get $hierarchy]
  set hier($id,top,fathers) ""
  set type_coord [position_of_type $hierarchy $type_id]
  set help [listOfSons $hierarchy $type_coord]
  set hier($id,$type_id,sons) [C_to_Tcllist $help]
  set childlist $hier($id,$type_id,sons)

  # recurse
  foreach value $childlist {
    compute_fathers_and_sons $hierarchy $value
  }
  foreach value $childlist {
    lappend hier($id,$value,fathers) $type_id
  }
}

## ----------------------------------------------------------------------------
## draw hierarchy on its associated canvas
## ----------------------------------------------------------------------------
body AllHierarchies::drawHierarchy {id} {

  # clear indices
  foreach x [array names widths] { unset widths($x) }
  foreach x [array names levels] { unset levels($x) }
  set maxlength 0
  set maxdepth 0

  # which canvas are we working on?
  set canvas $itk_component($id)

  # bind to event
  $canvas bind "label" <Any-Enter> [code $this labelEnter %W %X %Y]
  $canvas bind "label" <Any-Leave> [code $this labelLeave %W]

  # cdg's internal structure
  set h $hierarchies($id)

  # we want to re-arrange the tree for displaying, 
  # but xcdg shouldn't disturb cdg's data structures.
  # So we import the tree wholesale.
  importNode $id top

  if {$hier($id,is_a_tree)} {
    set_level_array $id top
    xpixels_per_character $id
    leaves_right $id top
    new_order $id top

    # print id and filename 
    # set id [HierarchyStruct_id_get $h]
    # regexp "\[^/\]*$"  [HierarchyStruct_filename_get $h] basename
    # $canvas create text 1 1 -anchor w -text "Hierarchy $id ($basename):"

    compute_positions $id
    drawSort $id top
  } else {
    set_level_array $id top
    xpixels_per_character $id
    compute_graph $id
  }
}

## ----------------------------------------------------------------------------
## fills global arrays with data from cdg.
## ----------------------------------------------------------------------------
body AllHierarchies::importNode {id type_id} {

  set l [string length $type_id]
  if {$l>$maxlength} {set maxlength $l}
  
  # check maximal tree depth
  set depth $hier($id,$type_id,max_type_depth)
  if { $depth > $maxdepth } { set maxdepth $depth }
  foreach value $hier($id,$type_id,sons) {
    importNode $id $value
  }
}

## ----------------------------------------------------------------------------
## populate the array levels.
## ----------------------------------------------------------------------------
body AllHierarchies::set_level_array {id type_id} {

  if {[string compare $type_id top] == 0} {
    for {set i 0} {$i <= $maxdepth} {incr i} {
      set levels($i) ""
    }
  }
  set level $hier($id,$type_id,max_type_depth)
  if {[lsearch $levels($level) $type_id] == -1} {
    lappend levels($level) $type_id
  }

  # recurse
  foreach value $hier($id,$type_id,sons) {
    set_level_array $id $value
  }
}

## ----------------------------------------------------------------------------
## rearrange children of a node so that wide nodes alternate with narrow nodes.
## ----------------------------------------------------------------------------
body AllHierarchies::balanceNode {id type_id} {

  # populate local arrays
  set i 0
  set min 99999
  set max 0
  foreach s $hier($id,$type_id,sons) {
    set desc($i) $s
    set w $hier($id,$s,number_sons)
    set widths($s) $w
    if {$w > $max} { set max $w }
    if {$w < $min} { set min $w }
    incr i
  }

  # populate global array levels
  set depth $hier($id,$type_id,max_type_depth)

  lappend levels($depth) $type_id

  # avoid unnecessary effort
  if { $max > $min } {

    # desc() := sort(desc())
    set maxi [expr $i-1]
    for { set i 0 } { $i <= $maxi } { incr i } {
      for { set j $i } { $j <= $maxi } { incr j } {
	set a $desc($i)
	set b $desc($j)
	if { $widths($a) < $widths($b)} {
	  set desc($i) $b
	  set desc($j) $a
	}
      }
    }

    # dsce() := shuffle(desc())
    for { set i 0 } { $i <= $maxi } { incr i } {
      if { [expr 2 * ($i / 2)] == $i } \
	  then { set j [expr $i / 2]} \
	  else { set j [expr $maxi - ($i/2)]}
      set dsce($i) $desc($j)
    }

    # register rearranged order of children
    set l ""
    for { set i 0 } { $i <= $maxi } { incr i } {
      append l " $dsce($i)"
    }
    set hier($id,$type_id,sons) $l
  }

  # recurse
  foreach value $hier($id,$type_id,sons) {
    balanceNode $id $value
  }
}

## ----------------------------------------------------------------------------
## draw one sort on the canvas.
## This method is only called when the hierarchy is a tree. 
## \note No check prevents us from calling this method on a non-tree hierarchy.
## ----------------------------------------------------------------------------
body AllHierarchies::drawSort {id type_id} {

  set levelheight 60

  set x $hier($id,$type_id,x_to)
  set x [expr 1.2 * $x]
  set hier($id,$type_id,x_to) $x

  set x $hier($id,$type_id,x_from)
  set x [expr 1.2 * $x]
  set hier($id,$type_id,x_from) $x

  set y [expr $levelheight * $hier($id,$type_id,max_type_depth)]

  # draw sort
  $canvas create text $x $y -font {Helvetica -12 bold} -text "$type_id" -anchor w

  # root?
  if {$hier($id,$type_id,max_type_depth) != 0} {

    # link to parent
    set parent $hier($id,$type_id,fathers)
    set x1 $hier($id,$parent,x_from)
    set y1 [expr $levelheight * $hier($id,$parent,max_type_depth)]
    set x [expr ( $hier($id,$type_id,x_from) + $hier($id,$type_id,x_to) ) / 2]
    set x1 [expr ( $hier($id,$parent,x_from) + $hier($id,$parent,x_to) ) / 2]
    $canvas create line $x [expr $y - 10]  $x1 [expr $y1 + 10]
  }

  # recurse
  foreach value $hier($id,$type_id,sons) {
    drawSort $id $value
  }
}

## ----------------------------------------------------------------------------
## calls initial_x_pos and center.
## ----------------------------------------------------------------------------
body AllHierarchies::compute_positions {id} {

  for {set i 0} {$i <= $maxdepth} {incr i} {
    initial_x_pos $id $i
  }
  for {set i 1} {$i <= $maxdepth} {incr i} {
    center $id $i
  }

  # Some nodes are still not positioned correctly.
  # start with lvl 1
  position_leaves $id 1

  # Leaves should be moved between their siblings
  # for better results:
  position_leaves2 $id top
}

## ----------------------------------------------------------------------------
## sort's initial position.
## ----------------------------------------------------------------------------
body AllHierarchies::initial_x_pos {id level} {

  set help $levels($level)
  set maxlistindex [expr [llength $help] - 1]
  set type [lindex $help 0]
  set hier($id,$type,x_from) 0
  set hier($id,$type,x_to) [expr $unit * ($hier($id,$type,stringlength) + 1)]
  set nextpos $hier($id,$type,x_to)
  for {set i 1} {$i <= $maxlistindex} {incr i} {
    set type [lindex $help $i]
    set hier($id,$type,x_from) $nextpos
    set hier($id,$type,x_to) \
      [expr $nextpos + $unit * ($hier($id,$type,stringlength) + 1)]
    set nextpos $hier($id,$type,x_to)
  }
}

## ----------------------------------------------------------------------------
## find sort's final x-coordinate.
## ----------------------------------------------------------------------------
body AllHierarchies::center {id level} {

  set help $levels($level)
  set fathers $levels([expr $level - 1])

  # Which indices have the sons?
  set index -1
  foreach father $fathers {
    if {$hier($id,$father,number_sons) == 0} {continue}
    set lastson($father) [expr $index + $hier($id,$father,number_sons)]
    set firstson($father) \
      [expr $lastson($father) + 1 - $hier($id,$father,number_sons)]
    set index $lastson($father)
  }

  # Which shift-operation is necessary? Will the predecessor be
  # shifted to the right above it`s successors, or the successors
  # to the right beneath their predecessor?
  foreach father $fathers {
    if {$hier($id,$father,number_sons) == 0} {continue}
    set fathers_index [lsearch -exact $fathers $father]
    set fathers_center \
      [expr ($hier($id,$father,x_from) + $hier($id,$father,x_to)) / 2]
    set leftson($father) [lindex $help $firstson($father)]
    set rightson($father) [lindex $help $lastson($father)]
    if {$leftson($father) == ""} {continue}
    if {$rightson($father) == ""} {
      set sons_center [expr ($hier($id,$leftson($father),x_from) \
        + $hier($id,$leftson($father),x_to)) / 2]
    } else {
      set sons_center [expr ($hier($id,$leftson($father),x_from) \
        + $hier($id,$rightson($father),x_to)) / 2]
    }
    if {$fathers_center < $sons_center} {

      # center fathers
      set recurse_necessary 1
      set new_x_from \
        [expr $sons_center - $unit * $hier($id,$father,stringlength) / 2]
      set how_far [expr $new_x_from - $hier($id,$father,x_from)]
      foreach father [lrange $fathers $fathers_index end] {
        set hier($id,$father,x_from) \
          [expr $hier($id,$father,x_from) + $how_far]
        set hier($id,$father,x_to) [expr $hier($id,$father,x_to) + $how_far]
      }
    } else {

      # center sons
      set how_far [expr ( $fathers_center - $sons_center ) / $unit ]
      set index [lsearch -exact $levels($level) $firstson($father)]
      foreach type [lrange $levels($level) $index end] {
	set hier($id,$type,x_from) [expr $hier($id,$type,x_from) + $how_far]
        set hier($id,$type,x_to) [expr $hier($id,$type,x_to) + $how_far]
      }
    }
  }
  set i [expr $level - 1]
  if {$recurse_necessary != 0} {
    if {$i > 0} {
      center $id $i
    }
  }
  set recurse_necessary 0
}

## ----------------------------------------------------------------------------
## move leaves to the rightmost possible position.
## Leaves are positioned right beside their non-leaf-siblings
## ----------------------------------------------------------------------------
body AllHierarchies::leaves_right {id type_id} {

  set help $hier($id,$type_id,sons)
  foreach son $help {
    if {$hier($id,$son,number_sons) == 0} {
      set hier($id,$type_id,sons) [ldelete $hier($id,$type_id,sons) $son]
      lappend hier($id,$type_id,sons) $son
    }
  }

  # recurse:
  foreach son $hier($id,$type_id,sons) {
    leaves_right $id $son
  }
}

## ----------------------------------------------------------------------------
## store the new order in 'levels(level)'
## ----------------------------------------------------------------------------
body AllHierarchies::new_order {id type_id} {

  set hier($id,top,max_type_depth) 0
  if {[string compare $type_id top] == 0} {
    for {set i 0} {$i <= $maxdepth} {incr i} {
      set levels($i) ""
    }
  }
  set level $hier($id,$type_id,max_type_depth)
  lappend levels($level) $type_id

  # recurse
  foreach value $hier($id,$type_id,sons) {
    new_order $id $value
  }
}

## ----------------------------------------------------------------------------
## deletes a string from a list.
## \todo this method is in no way specific to AllHierarchies
## ----------------------------------------------------------------------------
body AllHierarchies::ldelete { list type_id } {

  set ix [lsearch -exact $list $type_id]
  if {$ix >= 0} {
    return [lreplace $list $ix $ix]
  } else {
    return $list
  }
}

## ----------------------------------------------------------------------------
## some nodes are still not correctly positioned.
## ----------------------------------------------------------------------------
body AllHierarchies::position_leaves {id level} {

  set currentnode [lrange $levels($level) end end]
  set father $hier($id,$currentnode,fathers)
  set fathers_center \
    [expr ($hier($id,$father,x_from) + $hier($id,$father,x_to)) / 2]
  set leftson [lindex $hier($id,$father,sons) 0]
  set rightson [lindex $hier($id,$father,sons) end]
  set sons_center \
    [expr ($hier($id,$leftson,x_from) + $hier($id,$rightson,x_to)) / 2]
  if {$sons_center < $fathers_center} {
    set how_far [expr $fathers_center - $sons_center]
    foreach son $hier($id,$father,sons) {
      set hier($id,$son,x_from) [expr $hier($id,$son,x_from) + $how_far]
      set hier($id,$son,x_to) [expr $hier($id,$son,x_to) + $how_far]
    }
  }

  # recurse
  incr level
  if {$level <= $maxdepth} {
    position_leaves $id $level
  }
}

## ----------------------------------------------------------------------------
## position leaves between their siblings
## \todo bad method name
## ----------------------------------------------------------------------------
body AllHierarchies::position_leaves2 {id type_id} {

  # Find the leaf with greatest stringlength
  # Don't use this procedure with an empty list !

  proc leaf_max_stringlength {list_of_leaves} {; ## \type TclList
    upvar id id
    upvar hier hier
    set help [lindex $list_of_leaves 0]
    set leaf $help
    set stringlength $hier($id,$help,stringlength)
    foreach value $list_of_leaves {
      set help $hier($id,$value,stringlength)
      if {$help > $stringlength} {
        set stringlength $help
        set leaf $value
      }
    }
    return $leaf
  }

  # register leaves and non-leaves
  set leaves ""
  set nonleaves ""
  foreach son $hier($id,$type_id,sons) {
    if {$hier($id,$son,number_sons) == 0} {
      lappend leaves $son
    } else {
      lappend nonleaves $son
    }
  }

  # if leaves is empty, nothing is to be done except recurse
  # if the number of nonleaves is less than two, nothing is
  # to be done except recurse
  if {[llength $leaves] != 0 && [llength $nonleaves] > 1} {
    set help [lindex $leaves 0]

    # needed later
    set leaves_started_at $hier($id,$help,x_from)

    # register the spaces the leaves will be moved to as array new_home
    for {set i 0} {$i < [expr [llength $nonleaves] - 1]} {incr i} {
      set left [lindex $nonleaves $i]
      set right [lindex $nonleaves [expr $i + 1]]
      set from [expr $hier($id,$left,x_to) - $unit]
      set to $hier($id,$right,x_from)
      set remaining_space [expr $to - $from]
      set new_home($i,from) $from
      set new_home($i,to) $to
      set new_home($i,remaining_space) $remaining_space
      set new_home($i,nodes) ""
      set lastindex $i
    }

    # enforce the first iteration:
    set something_changed 1
    while {$something_changed == 1} {
      set something_changed 0

      # find the largest  space
      set index_largest_space 0
      set space $new_home(0,remaining_space)
      for {set i 1} {$i <= $lastindex} {incr i} {
        set help $new_home($i,remaining_space)
        if {$help > $space} {
          set space $help
          set index_largest_space $i
        }
      }

      # index_largest_space is the index of the space to work with
      set helplist $leaves
      while {$helplist != ""} {
        set node [leaf_max_stringlength $helplist]
        set helplist [ldelete $helplist $node]
        set need [expr $hier($id,$node,x_to) - $hier($id,$node,x_from)]
        set remaining $new_home($index_largest_space,remaining_space)
        if {$need <= $remaining} {
          set new_home($index_largest_space,remaining_space) \
           [expr $remaining - $need]
          lappend new_home($index_largest_space,nodes) $node
          set leaves [ldelete $leaves $node]
          set something_changed 1
        }
      }
    }

    # find the leaves' new positions
    foreach value $leaves {
      set howfar [expr $leaves_started_at - $hier($id,$value,x_from)]
      set hier($id,$value,x_from) $leaves_started_at
      set hier($id,$value,x_to) [expr $hier($id,$value,x_to) + $howfar]
      set leaves_started_at $hier($id,$value,x_to)
    }
    for {set i 0} {$i <= $lastindex} {incr i} {
      set number_of_leaves [llength $new_home($i,nodes)]
      set available_space [expr $new_home($i,to) - $new_home($i,from)]
      foreach value $new_home($i,nodes) {
        set available_space \
          [expr $available_space - $hier($id,$value,stringlength) * $unit]
      }
      set between [expr floor($available_space / ($number_of_leaves + 1))]
      set starting_position [expr $new_home($i,from) + $between]
      foreach value $new_home($i,nodes) {
        set howfar [expr $starting_position - $hier($id,$value,x_from)]
	set hier($id,$value,x_from) [expr $hier($id,$value,x_from) + $howfar]
        set hier($id,$value,x_to) [expr $hier($id,$value,x_to) + $howfar]
        set starting_position [expr $hier($id,$value,x_to) + $between]
      }
    }
  }

  # recurse
  foreach value $hier($id,$type_id,sons) {
    position_leaves2 $id $value
  }
}

## ----------------------------------------------------------------------------
## This function is used only for nontree-graphs.
## ----------------------------------------------------------------------------
body AllHierarchies::compute_graph {id} {

  set yunit 16
  set xunit [expr 2 * $unit]

  # can range from 0 (one node is allways printed) to 38
  set maxnodes 34

  for {set level 0} {$level < $maxdepth} {incr level} {
    set counter 0
    set help $levels($level)
    set first_iteration 1
    while {$help != ""} {
      set fathers ""
      set nodes_current_lvl ""
      set sons ""
      set leaves ""
      foreach value $help {
        if {$hier($id,$value,number_sons) == 0} {
          lappend leaves $value
        } else {
          lappend fathers $value
        }
      }
      set lonely_father [lindex $fathers 0]
      set maxindex [expr [llength $fathers] - 1]
      for {set index 0} {$index <= $maxindex} {incr index} {
        set father [lindex $fathers $index]
        set count 0
        foreach son $hier($id,$father,sons) {
          if {[lsearch $sons $son] == -1} {
            lappend sons $son
            incr count
          }
        }
        set helpint \
          [expr 1 + $count + [llength $nodes_current_lvl] + [llength $sons]]
        if {$helpint <= $maxnodes} {
          lappend nodes_current_lvl $father
          set help [ldelete $help $father]
        }
      }

      if {[llength $nodes_current_lvl] == 0} {

        # No father was attached to list nodes_current_lvl, because it
        # wouldn't fit on the screen with all of it's direct successors.
        # Nevertheless, at least one node should be printed:
        lappend nodes_current_lvl $lonely_father
	set help [ldelete $help $lonely_father]
      }

      # Fathers that don't introduce new sons can still be attached to list
      # nodes_current_lvl.
      set max_additional_nodes [llength $sons]
      while {$max_additional_nodes > 0 && $index <= $maxindex} {
        set father [lindex $fathers $index]
        set attach 1
        foreach son $hier($id,$father,sons) {
	  if {[lsearch $sons $son] == -1} {
            set attach 0
          }
        }
        if {$attach == 1} {
          lappend nodes_current_lvl $father
          set help [ldelete $help $father]
	  set max_additional_nodes [expr $max_additional_nodes - 1]
        }
        incr index
      }

      # Put the leaves at the end of the first column:
      foreach leaf $leaves {
        lappend nodes_current_lvl $leaf
        set help [ldelete $help $leaf]
      }
      if {$first_iteration == 1} {
        $canvas create text 1 $current_y -anchor w \
           -text "Level $level and direct successors"
        set current_y [expr $current_y + 2 * $yunit]
      }
      display_graph \
        $id $level $nodes_current_lvl $current_x $current_y $counter
      set first_iteration 0
      incr counter
    }
    set current_y [expr $max_y + 2 * $yunit]
    set current_x 1
  }
}

## ----------------------------------------------------------------------------
## display nontree-graphs.
## This function is used only for nontree-graphs
## ----------------------------------------------------------------------------
body AllHierarchies::display_graph {id level nodes_current_lvl x_start y_start counter} {

  upvar yunit yunit
  upvar xunit xunit
  set fathers ""
  set leaves ""
  set sons ""

  # Ugly patch. Should somehow be removed later:
  foreach value $nodes_current_lvl {
    if {$hier($id,$value,number_sons) == 0} {
      lappend leaves $value
    } else {
      lappend fathers $value
    }
  }

  # assemble list of sons:
  foreach father $fathers {
    foreach son $hier($id,$father,sons) {
      if {[lsearch -exact $sons $son] == -1} {
        lappend sons $son
      }
    }
  }

  # vertical lines:
  set draw_to_y \
    [expr $current_y + ([llength $fathers] + [llength $sons] - 1) * $yunit]

  # Needed later:
  set keep_y $y_start

  set max_x 0

  # Needed for the horizontal lines:
  set index 0

  # sons' starting-position:
  set sons_start 0

  # Write this levels' nodes (except for the leaves) to the canvas:
  foreach node $fathers {

    # $level and $counter are appended to the labels to make the
    # different graphs on the canvas independent of each other
    set fatherstag father_
    append fatherstag $node $level $counter

    # build list of sons
    set list_o_sons ""
    foreach son $hier($id,$node,sons) {
      set sonstag son_
      append sonstag $son $level $counter
      if {[lsearch -exact $list_o_sons $sonstag] == -1} {
        lappend list_o_sons $sonstag
      }
    }

    set father_taglist [linsert $list_o_sons 0 $fatherstag]
            
    $canvas create text $current_x $current_y -anchor w \
       -text "$node" -tags [linsert $father_taglist 1 "label"]
    set help [expr $current_x + $hier($id,$node,stringlength) * $unit]


    if {$help > $max_x} {
      set max_x $help
    }

    # create this nodes horizontal and vertical lines (if it is no leaf):
    if {$hier($id,$node,number_sons) != 0} {
      set draw_from_x [expr ($hier($id,$node,stringlength) + 2) \
        * $unit + $x_start - 5]
      set draw_to_x \
        [expr $maxlength * $unit + $x_start + ($index + 1) * $xunit]

      # The first fathers draw_to_x is needed for the sons' horizontal lines
      if {[lsearch $fathers $node] == 0} {
        set sons_draw_from_x $draw_to_x
      }

      # horizontal line:
      $canvas create line $draw_from_x $current_y $draw_to_x $current_y \
         -tags $father_taglist

      # starting-position for the sons:
      if {$draw_to_x > $sons_start} {
        set sons_start $draw_to_x
      }

      # vertical line:
      $canvas create line $draw_to_x $current_y $draw_to_x $draw_to_y \
         -tags $father_taglist

      # store information for the drawing and tagging of the dots:
      set x($node) $draw_to_x
      set dot($node,x) $draw_to_x
    }
    set current_y [expr $current_y + $yunit]
    incr index
  }
  set sons_x [expr $sons_start + 20]
  set sons_y_start_at $current_y

  # write the leaves to the canvas:
  foreach node $leaves {
    $canvas create text $x_start $current_y -anchor w -text "$node"
    set help [expr $x_start + $hier($id,$node,stringlength) * $unit]
    if {$help > $max_x} {
      set max_x $help
    }
    set current_y [expr $current_y + $yunit]
    if {$current_y > $max_y} {
      set max_y $current_y
    }
  }
  set current_y $sons_y_start_at
  for {set i $level} {$i <= $maxdepth} {incr i} {  
    set offset [expr ($i - $level) * 20]

    # write the sons to the canvas
    foreach node $sons {
      set sonstag son_
      append sonstag $node $level $counter

      if {$hier($id,$node,max_type_depth) == $i} {
        
	# build list of fathers
	set list_o_fathers ""
	foreach father $hier($id,$node,fathers) {
          set fatherstag father_
          append fatherstag $father $level $counter
          if {$hier($id,$father,max_type_depth) == $level} {
            if {[lsearch -exact $list_o_fathers $fatherstag] == -1} {
	      lappend list_o_fathers $fatherstag
            } 
          }
        }
        
	set sons_taglist [linsert $list_o_fathers 0 $sonstag]
		
	$canvas create text \
          [expr $sons_x + $offset] $current_y -anchor w -text "$node" \
          -tags [linsert $sons_taglist 1 "label"]

        $canvas create line $sons_draw_from_x $current_y \
          [expr $sons_x + $offset - 10] $current_y \
	  -tags $sons_taglist
        
        # store information for the drawing of the dots:
        set dot($node,y) $current_y
        set help \
          [expr $sons_x + $offset + $hier($id,$node,stringlength) * $unit]
        if {$help > $max_x} {
          set max_x $help
        }
        set current_y [expr $current_y + $yunit]
        set sons [ldelete $sons $node]
        if {$current_y > $max_y} {
          set max_y $current_y
        }
      }
    }
  }

  # print and label the dots:
  foreach father $fathers {
    foreach son $hier($id,$father,sons) {
      set fatherstag father_
      append fatherstag $father $level $counter
      set sonstag son_
      append sonstag $son $level $counter
      set dotx $dot($father,x)
      set doty $dot($son,y)
      $canvas create oval [expr $dotx - 2] [expr $doty - 2] \
         [expr $dotx + 2] [expr $doty + 2] \
         -fill black -width 0 -tags [list $fatherstag $sonstag]
    }
  }
  set current_x [expr $max_x + 50]
  set current_y $keep_y
}

## ----------------------------------------------------------------------------
## How many pixels are to be used for a character ?
## ----------------------------------------------------------------------------
body AllHierarchies::xpixels_per_character {id} {

  set teststring [lindex $levels(1) 0]
  set length $hier($id,$teststring,stringlength)
  $canvas create text 2 2 -anchor w -text "$teststring" -tag remove
  set help [$canvas bbox remove]
  scan $help "%f %f %f %f" x1 y1 x2 y2
  $canvas delete remove
  set help [expr ($x2 - $x1) / $length]
  set unit [expr ceil($help)]
}

## ----------------------------------------------------------------------------
## command bound to Enter-event.
## ----------------------------------------------------------------------------
body AllHierarchies::labelEnter {w pointerx pointery} {

  set tags [$w gettags current]
  set relevant_label [lindex $tags 0]
  $w itemconfigure $relevant_label -fill red
}

## ----------------------------------------------------------------------------
## command bound to Leave-event.
## ----------------------------------------------------------------------------
body AllHierarchies::labelLeave {w} {

  set tags [$w gettags current]
  set relevant_label [lindex $tags 0]
  $w itemconfigure $relevant_label -fill black
}

## ----------------------------------------------------------------------------
## Return list of all hierarchies' names
## ----------------------------------------------------------------------------
body AllHierarchies::getAllHierarchyIds {} {
  init_data
  return [array names hierarchies]
}
