# Copyright (C) 1997-2005 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read COPYING in the root of this distribution.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaGrammars - configure document for all grammars.
## \ingroup YadaConfiguration
##
## \author Michael Daum
##
## ----------------------------------------------------------------------------
class YadaGrammars {
  inherit YadaConfigDocument


  # variables ----------------------------------------------------------------
  private variable _noAllWordgraphs 0
  private variable _noSelWordgraphs 0

  # methods ------------------------------------------------------------------
  public method init {}
  public method show {}

  constructor {args} {}; ## \type TclList


  protected method isValidPersistenceFile {fileName}; ## \type TclString

  private method _select {{grammarName ""}}; ## \type TclString
  private method _commit {}
  private method _new {}
  private method _delete {}
  private method _editGrammar {}
  private method _fileChooser {}
  private method _mark {aspect}; ## \type TclString
  private method _getSelectedWordgraphs {}
  private method _selectWordgraphs {{wordgraphs ""}}; ## \type TclList
  private method _unselectWordgraphs {{wordgraphs ""}}; ## \type TclList
  private method _invertWordgraphs {{wordgraphs ""}}; ## \type TclList
  private method _defaults {}
  private method _leaveTableCommand {}
  private method _escapeCommand {}
  private method _returnCommand {}
  private method _randomSelection {}
  private method _lengthSelection {{min -1} {max -1}} ;## \type TclNumber, TclNumber
  private method _validateCommand {aspect newValue} ;## \type TclString, TclString
  private method _rescanGrammar {}
  private method _setCell {args}; ## \type TclList
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body YadaGrammars::constructor {args} {

  set _allAspects {name files wordgraphs}

  # edit menu
  set index [$itk_component(editMenu) index "Delete"]
  incr index
  $itk_component(editMenu) insert [incr index] command \
       	-label "Edit Grammar" \
       	-underline 0 \
       	-command [code $this _editGrammar]
  $itk_component(editMenu) insert [incr index] command \
       	-label "Rescan" \
       	-underline 0 \
       	-command [code $this _rescanGrammar]
  $itk_component(editMenu) insert [incr index] separator
  $itk_component(editMenu) insert [incr index] command \
       	-label "Select" \
       	-underline 0 \
       	-command [code $this _selectWordgraphs]
  $itk_component(editMenu) insert [incr index] command \
       	-label "Unselect" \
       	-underline 0 \
       	-command [code $this _unselectWordgraphs]
  $itk_component(editMenu) insert [incr index] command \
       	-label "Invert Selection" \
       	-underline 0 \
       	-command [code $this _invertWordgraphs]
  $itk_component(editMenu) insert [incr index] command \
       	-label "Random Selection" \
       	-underline 0 \
       	-command [code $this _randomSelection]
  $itk_component(editMenu) insert [incr index] command \
       	-label "Length Selection" \
       	-underline 0 \
       	-command [code $this _lengthSelection]
  $itk_component(editMenu) insert [incr index] separator

  # more buttons
  itk_component add editButton {
    button $itk_component(topButtonFrame).editButton \
	-text "Edit" \
	-command [code $this _editGrammar]
  }

  itk_component add rescanButton {
    button $itk_component(topButtonFrame).rescanButton \
	-text "Rescan" \
	-command [code $this _rescanGrammar]
  }

  itk_component add selButton {
    button $itk_component(topButtonFrame).selButton \
      -text "Select" \
      -command [code $this _selectWordgraphs]
  }

  itk_component add unselButton {
    button $itk_component(topButtonFrame).unselButton \
      -text "Unselect" \
      -command [code $this _unselectWordgraphs]
  }


  itk_component add invertButton {
    button $itk_component(topButtonFrame).invertButton \
      -text "Invert" \
      -command [code $this _invertWordgraphs]
  }
  itk_component add randomButton {
    button $itk_component(topButtonFrame).randomButton \
      -text "Random" \
      -command [code $this _randomSelection]
  }
  itk_component add lengthButton {
    button $itk_component(topButtonFrame).lengthButton \
      -text "Length" \
      -command [code $this _lengthSelection]
  }

  #
  # files 
  itk_component add filesLabel {
    label $itk_component(childsite).filesLabel \
      -text "Files:" \
      -anchor nw 
  } {}

  itk_component add filesMark {
    frame $itk_component(childsite).filesMark \
      -borderwidth 2
  } {}

  itk_component add files {
    entry $itk_component(filesMark).files \
      -borderwidth 2 \
      -validate all \
      -validatecommand [code $this _validateCommand "files" %P]
  } {}

  itk_component add filesButton {
    button $itk_component(filesMark).filesButton \
      -image [YadaImages::get defaultFolder] \
      -padx 5 \
      -pady 0 \
      -command [code $this _fileChooser]
  } {}

  #
  # the wordgraph selector
  #
  itk_component add wordgraphsMark {
    frame $itk_component(childsite).wordgraphsMark \
      -borderwidth 2
  } {}

  itk_component add wordgraphLabel {
    label $itk_component(wordgraphsMark).label \
      -text "Wordgraphs:" \
      -anchor nw
  } {}

  itk_component add counterFrame {
    frame $itk_component(wordgraphsMark).counterFrame 
  } {}
  itk_component add totalCount {
    iwidgets::entryfield $itk_component(counterFrame).totalCount \
      -labeltext "Total:"\
      -labelpos w \
      -state disabled \
      -textvariable [scope _noAllWordgraphs] \
      -borderwidth 2 \
      -disabledforeground black \
  } {}

  itk_component add selCount {
    iwidgets::entryfield $itk_component(counterFrame).selCount \
      -labeltext "Selected:"\
      -labelpos w \
      -state disabled \
      -textvariable [scope _noSelWordgraphs] \
      -borderwidth 2 \
      -disabledforeground black \
  } {}

  itk_component add table {
    MyTable $itk_component(wordgraphsMark).table \
      -exportselection 1 \
      -selecttitles 1 \
      -rowseparator "\n" \
      -colseparator "\t" \
      -rows 20 \
      -cols 5 \
      -state disabled \
      -cellanchor w \
      -colstretchmode last \
      -rowstretchmode none \
      -selectmode extended \
      -selecttype row \
      -titlerows 1 \
      -titlecols 1 \
      -borderwidth 3 \
      -outerborderwidth 2 \
      -outerrelief sunk \
      -selforeground "" \
      -selbackground "" \
      -font {-adobe-helvetica-medium-r-normal--12-*-*-*-*-*-*-*} \
      -hscrollmode static \
      -vscrollmode static \
  } {}

  _setCell \
      0,0 "no." \
      0,1 "wordgraph" \
      0,2 "length" \
      0,3 "trivial" \
      0,4 "sentence"

  $itk_component(table) tag config title \
    -font {-adobe-helvetica-medium-r-normal--12-*-*-*-*-*-*-*} -anchor c 
  $itk_component(table) tag config sel -fg "" -bg "" -relief raised -bd 1
  $itk_component(table) tag config mark -bg green3 -fg black
  $itk_component(table) width 0 4 1 15 2 7 3 5 

  #
  # packing + gridding
  #
  grid rowconfigure $itk_component(childsite) 3 -weight 1

  pack $itk_component(editButton) \
       $itk_component(rescanButton) \
       $itk_component(selButton) \
       $itk_component(unselButton) \
       $itk_component(invertButton) \
       $itk_component(randomButton) \
       $itk_component(lengthButton) -side top -fill x -pady 5

  grid $itk_component(filesLabel) -sticky ew -row 1 -column 0 -pady 10
  grid $itk_component(filesMark) -sticky ew -row 1 -column 1 -pady 10
  pack $itk_component(files) -fill both -expand 1 -side left
  pack $itk_component(filesButton) -side left

  grid $itk_component(wordgraphsMark) -sticky news -row 3 -column 0 -pady 10 -columnspan 3
  grid columnconfigure $itk_component(wordgraphsMark) 0 -weight 1
  grid rowconfigure $itk_component(wordgraphsMark) 1 -weight 1

  grid $itk_component(wordgraphLabel) -sticky nw -row 0 -column 0 
  grid $itk_component(table)          -sticky news -row 1 -column 0 -columnspan 2 -pady 5
  grid $itk_component(counterFrame)   -sticky ew -row 2 -column 0 -columnspan 2 -pady 5
  pack $itk_component(totalCount) \
       $itk_component(selCount) -side left


  # bindings
  bind [$itk_component(table) component table] <Double-Button-1> \
    +[code $this _invertWordgraphs]

  set hull $itk_component(hull)
  set table $itk_component(table)
  set tableInTable [$table component table]
  bind $hull <Leave> [code $this _leaveTableCommand]
  bind $hull <FocusOut> [code $this _leaveTableCommand]
  bind $tableInTable <KeyPress-Escape> [code "$this _escapeCommand; break"]
  bind $tableInTable <KeyPress-Return> [code "$this _returnCommand; break"]

  eval itk_initialize $args
}

## ----------------------------------------------------------------------------
## init
## ----------------------------------------------------------------------------
body YadaGrammars::init {} {
  if {$_isInitialized} {
    return
  }

  # get xml data
  set found 0
  set prefs [[.main component Configure] component prefs]
  foreach itemRefNode [$prefs evalXpath "/yada/global/$name/*"] {
    set itemName [$itemRefNode getAttribute "name"]
    set itemFileName [$itemRefNode getAttribute "fileName"]
    if {[file exists $itemFileName]} {
      set item [newGrammar -name $itemName]
      $item load $itemFileName
      if {[addItem $item]} {
	incr found
      } else {
	itcl::delete object $item
      }
    }
  }

  if {!$found} {
    _defaults
  }

  # required <none> grammar
  addItem [newGrammar -name "<none>"]

  chain

  _selectFirst
}


## ----------------------------------------------------------------------------
## _select
## ----------------------------------------------------------------------------
body YadaGrammars::_select {{grammarName ""}} {

  # get a grammar name from the combobox if we didn't provide one
  if {$grammarName == ""} {
    set grammarName [$itk_component(name) get]
  } 

  set item [getItem $grammarName]
  if {$item == ""} {
    printMessage "ERROR: grammar '$grammarName' is unknown" error
    return
  }

  # check if we didn't commit changes
  if {![_askChange]} {
      return
  }

  # set the active grammar
  setCurrentItem $item

  # update title bar and menus
  updateWindowTitle
  updateMenus

  # update the name box
  set box $itk_component(name)
  set grammarNames [[$box component list] get 0 end]
  set index [lsearch $grammarNames $grammarName]
  $box selection set $index $index
  $box delete entry 0 end
  $box insert entry 0 $grammarName

  # update the entries
  set entry $itk_component(files) 
  $entry delete 0 end
  $entry insert 0 [$item cget -files]

  # remove marking
  _unmark

  # show the wordgraphs
  show

}

## ----------------------------------------------------------------------------
## _commit
## ----------------------------------------------------------------------------
body YadaGrammars::_commit {} {

  set currentItem [getCurrentItem]
  set oldName [$currentItem cget -name]
  set oldFiles [$currentItem cget -files]
  set oldSelection [$currentItem getSelection]

  if {![chain]} {
    return
  }

  # adjust each experiment using a grammar with a changed name
  set newName [$currentItem cget -name]
  if {$oldName != $newName} {
    foreach experiment [.main getExperiments] {
      if {[$experiment cget -grammarName] == $oldName} {
	$experiment configure -grammarName $newName
      }
    }
  }
  
  # save the files
  $currentItem configure -files [$itk_component(files) get]

  # save the wordgraph selection
  eval $currentItem setSelection [_getSelectedWordgraphs]

  set newFiles [$currentItem cget -files]
  set newSelection [$currentItem getSelection]

  # recheck every experiment using this grammar
  if {$oldFiles != $newFiles || $oldSelection != $newSelection} {
    foreach experiment [.main getExperiments] {

      if {[$experiment cget -grammarName] == $newName} {
	.main printStatus "... checking experiment [$experiment cget -name]"
	$experiment configure -isUpToDate 0
	$experiment check
      }
    }
    .main printStatus ""
  }

  # remove marking
  _unmark
  _select
}

## ----------------------------------------------------------------------------
## _delete
## ----------------------------------------------------------------------------
body YadaGrammars::_delete {} {

  # get the grammar name
  set currentItem [getCurrentItem]
  set grammarName [$currentItem cget -name]

  # grammar "<none>" is undeleteable
  if {$grammarName == "<none>"} {
    return
  }

  # set experiments using the deleted grammar to grammar <none>
  set noExperiments 0
  foreach experiment [.main getExperiments] {
    if {[$experiment cget -grammarName] == $grammarName} {
      incr noExperiments
    }
  }

  set msgText "Are you sure that you want to delete grammar '$grammarName'?"
  if {$noExperiments > 0} {
    if {$noExperiments == 1} {
      append msgText "\nThere is one experiment using it."
    } else {
      append msgText "\nThere are $noExperiments experiments using it."
    }
  }

  # are you sure?
  set msg [iwidgets::messagedialog .msg \
    -text $msgText \
    -title "Yada - Caution"  \
    -bitmap questhead \
    -modality application \
    -master $itk_interior]
  $msg buttonconfigure OK -text "Yes"
  $msg buttonconfigure Cancel -text "No"
  $msg center .
  set result [$msg activate]
  destroy $msg
  if {!$result} {
    return
  }

  # delete the data
  removeItem $currentItem
  $currentItem delete
  itcl::delete object $currentItem

  # update the names
  updateCombos

  # set experiments using the deleted grammar to grammar <none>
  foreach experiment [.main getExperiments] {
    if {[$experiment cget -grammarName] == $grammarName} {
      $experiment configure -grammarName "<none>" -isModified 1
    }
  }

  # show something
  _selectFirst
}

## ----------------------------------------------------------------------------
## _new
## ----------------------------------------------------------------------------
body YadaGrammars::_new {} {

  # ask for a new name
  set prd [iwidgets::promptdialog .prd \
    -title "Yada - New grammar" \
    -modality application \
    -labeltext "Name:" \
    -labelpos w]

  set grammarName [$itk_component(name) get]
  set entry [$prd component prompt]
  $entry configure -width 40
  $entry insert 0 $grammarName

  $prd center .
  $prd hide Apply
  $prd hide Help
  after idle "focus [$prd component prompt]"

  set result [$prd activate]
  set grammarName [$prd get]
  destroy $prd
  
  if {!$result} {
    return
  }

  set item [getItem $grammarName]
  if {$item != ""} {
    printMessage "ERROR: grammar '$grammarName' exists already." error
    return [_new]
  }

  # build a new grammar
  set currentItem [getCurrentItem]
  setCurrentItem [$currentItem clone -name $grammarName]
  
  # store all aspects
  set box $itk_component(name)
  $box delete entry 0 end
  $box insert entry 0 $grammarName
  _commit

  # display it 
  _select $grammarName
}

## ----------------------------------------------------------------------------
## _validateCommand
## ----------------------------------------------------------------------------
body YadaGrammars::_validateCommand {aspect newValue} {

  switch $aspect {
    "wordgraphs" {
      set currentItem [getCurrentItem]
      set oldValue [$currentItem getSelection]
    }
    "files" {
      set currentItem [getCurrentItem]
      set oldValue [string trim [$currentItem cget -files]]
    }
    default {
      set oldValue ""
    }
  }

  chain $aspect $newValue $oldValue

  return 1
}

## ----------------------------------------------------------------------------
## _mark
## ----------------------------------------------------------------------------
body YadaGrammars::_mark {aspect} {

  switch $aspect {
    "wordgraphs" {
      set currentItem [getCurrentItem]
      set newData [$this _getSelectedWordgraphs]
      set oldData [$currentItem getSelection]
    }
    "files" {
      set currentItem [getCurrentItem]
      set newData [$itk_component($aspect) get]
      set oldData [string trim [$currentItem cget -files]]
    }
    default {
      set newData ""
      set oldData ""
    }
  }

  chain $aspect $newData $oldData
}

## ----------------------------------------------------------------------------
## show
## ----------------------------------------------------------------------------
body YadaGrammars::show {} {

  # clear the tables
  set table $itk_component(table)
  set rmax [$table cget -rows]
  $table erase 1,1 $rmax,4
  $table clear tags 1,1 $rmax,4
  $table selection clear all

  # get all wordgraphs
  set currentItem [getCurrentItem]
  set currentName [$currentItem cget -name]
  set isScanned [$currentItem isScanned]
  set allWordgraphs [$currentItem getWordgraphs]

  if {[isVisible] && $currentName != "<none>"} {

    # get a reason why to rescan the current grammar
    set reason ""

    # if there are no wordgraphs scan for them if we didn't so before
    if {$allWordgraphs == "" && !$isScanned} {
      set reason "No wordgraphs selected!"
    }
      
    # do the dialog if there's a reason
    if {$reason != ""} {
      set msg [iwidgets::messagedialog .msg \
	       -text "$reason Rescan grammar '$currentName'?" \
	       -title "Yada - Question"  \
	       -image [YadaImages::get question] \
	       -modality application \
	       -master $itk_interior]

      $msg buttonconfigure OK -text "OK"
      $msg buttonconfigure Cancel -text "Cancel"
      $msg center .
      set result [$msg activate]
      destroy $msg
      if {$result} {
	_rescanGrammar
	set allWordgraphs [$currentItem getWordgraphs]
      }
    }
  }

  
  # set the total counter
  set _noAllWordgraphs [llength $allWordgraphs]
  set _noSelWordgraphs 0
  if {$_noAllWordgraphs < 20} {
    set rmax 20
  } else {
    set rmax [expr $_noAllWordgraphs +1]
  }
  $table configure -rows $rmax

  # insert a row for each wordgraph
  set row 1
  foreach wordgraph $allWordgraphs {
    set words [$wordgraph getWords]
    _setCell \
      $row,0 [format "%4d" $row ] \
      $row,1 [$wordgraph cget -name] \
      $row,2 [llength $words] \
      $row,3 [$wordgraph cget -isTrivial] \
      $row,4 [join $words]

    # establish the selection
    set wgName [$wordgraph cget -name]
    if {[$currentItem isSelected $wgName]} {
      $table tag row mark $row
      incr _noSelWordgraphs
    } else {
      $table tag row "" $row
    }
    incr row
  }

  # adjust the view
  $table see 1,1
}

## ----------------------------------------------------------------------------
## _selectWordgraphs
## ----------------------------------------------------------------------------
body YadaGrammars::_selectWordgraphs {{wordgraphs ""}} {

  set table $itk_component(table)

  if {$wordgraphs == ""} {
    # get selection from table
    foreach cell [$table curselection] {
      scan $cell "%d,%d" row col
      if {$row > 0 && $col == 1} {
	if {[$table getCell $row,0] != ""} {
	  if {[$table selection includes $cell]} {
	    $table tag row mark $row
	  } 
	}
      }
    }
  } else {
    # select given wordgraphs 
    set rmax [$table cget -rows]
    set ids [$table getCell 0,1 $rmax,1]
    foreach wordgraph $wordgraphs {
      set row [lsearch -exact $ids $wordgraph]
      $table tag row mark $row
    }
  }
  
  set _noSelWordgraphs [llength [_getSelectedWordgraphs]]
  _mark "wordgraphs"
}

## ----------------------------------------------------------------------------
## _unselectWordgraphs
## ----------------------------------------------------------------------------
body YadaGrammars::_unselectWordgraphs {{wordgraphs ""}} {
  
  set table $itk_component(table)

  if {$wordgraphs == ""} {
    # get selection from table
    foreach cell [$table curselection] {
      scan $cell "%d,%d" row col
      if {$row > 0 && $col == 1} {
	if {[$table getCell $row,0] != ""} {
	  if {[$table selection includes $cell]} {
	    $table tag row "" $row
	  } 
	}
      }
    }
  } else {
    # unselect given wordgraphs 
    set rmax [$table cget -rows]
    set ids [$table getCell 0,1 $rmax,1]
    foreach wordgraph $wordgraphs {
      set row [lsearch -exact $ids $wordgraph]
      $table tag row "" $row
    }
  }

  set _noSelWordgraphs [llength [_getSelectedWordgraphs]]
  _mark "wordgraphs"
}

## ----------------------------------------------------------------------------
## _invertWordgraphs
## ----------------------------------------------------------------------------
body YadaGrammars::_invertWordgraphs {{wordgraphs ""}} {

  set table $itk_component(table)

  if {$wordgraphs == ""} { 
    # get selection from table
    foreach cell [$table curselection] {
      scan $cell "%d,%d" row col
      if {$row > 0 && $col == 1} {
	if {[$table getCell $row,0] != ""} {
	  if {[$table selection includes $cell]} {
	    if {[$table tag includes mark $cell]} {
	      $table tag row "" $row
	    } else {
	      $table tag row mark $row
	    }
	  } 
	}
      }
    }
  } else {
    # invert given wordgraphs 
    set rmax [$table cget -rows]
    set ids [$table getCell 0,1 $rmax,1]
    incr row
    foreach wordgraph $wordgraphs {
      set row [lsearch -exact $ids $wordgraph]
      if {[$table tag includes mark $row,1]} {
	$table tag row "" $row
      } else {
	$table tag row mark $row
      }
    }
  }
  
  set _noSelWordgraphs [llength [_getSelectedWordgraphs]]
  _mark "wordgraphs"
}

## ----------------------------------------------------------------------------
## _editGrammar
## ----------------------------------------------------------------------------
body YadaGrammars::_editGrammar {} {
  set editor [.main getPreference editor]
  set currentItem [getCurrentItem]
  catch {eval exec -- $editor [$currentItem cget -files]}
}

## ----------------------------------------------------------------------------
## create YadaGrammar objects sourcing a tcl resource file.
## The resource file is either @c YADA_ETC/grammars.tcl if it exists or
## @c YADA_LIB/grammars.tcl.sample. The former allows the user to programm
## grammars. 
## ----------------------------------------------------------------------------
body YadaGrammars::_defaults {} {
  global env

  # get default resources
  set fileName [file join $::env(YADA_ETC) grammars.tcl]
  if {[file exists $fileName]} {
    .main printStatus "... loading $name file '$fileName'."
  } else {
    .main printStatus "... configure file '$fileName' not found ... using defaults."
    set fileName [file join $env(YADA_LIB) grammars.tcl.sample]
  }

  # ... configured grammars
  if {[catch {source $fileName} errMsg]} {
    bgerror "There's an error in your grammars init file $fileName!"
  }

  # assert grammars into the class
  set n 0
  foreach grammar [find objects -class YadaGrammar] {
    if {[addItem $grammar]} {
      incr n
    } else {
      itcl::delete object $grammar
    }
  }

  .main printStatus "... built $n grammar(s)"
}

## ----------------------------------------------------------------------------
## _leaveTableCommand
## ----------------------------------------------------------------------------
body YadaGrammars::_leaveTableCommand {} {
  set table $itk_component(table)
  set lastActiveCell [$table tag cell active]
  if {$lastActiveCell == ""} {
    return
  }

  scan $lastActiveCell "%d,%d" row col
  $table activate -1,-1
}

## ----------------------------------------------------------------------------
## _returnCommand
## ----------------------------------------------------------------------------
body YadaGrammars::_returnCommand {} {
  set table $itk_component(table)
  set lastActiveCell [$table tag cell active]
  scan $lastActiveCell "%d,%d" row col
  set colMax [expr [$table cget -cols] -1]
  if {$col == $colMax} {
    incr row
    set col 1
  } else {
    incr col
  }

  $table activate $row,$col
  $table see $row,$col
}

## ----------------------------------------------------------------------------
## _escapeCommand
## ----------------------------------------------------------------------------
body YadaGrammars::_escapeCommand {} {
  set table $itk_component(table)
  set lastActiveCell [$table tag cell active]
  scan $lastActiveCell "%d,%d" row col
  
  if {$col == 0} {
    $table selection clear all
  } else {
    $table activate -1,-1
  }
}

## ----------------------------------------------------------------------------
## select a random subset of the already selected wordgraphs.
## ----------------------------------------------------------------------------
body YadaGrammars::_randomSelection {} {
  set grammarName [$itk_component(name) get]
  if {$grammarName == "<none>"} {
    return
  }

  # get current selected wordgraphs
  set allWordgraphNames [_getSelectedWordgraphs]

  # unselect all
  set table $itk_component(table)
  set rmax [$table cget -rows]
  for {set row 0} {$row < $rmax} {incr row} { 
    $table tag row "" $row
  }

  # choose some random wordgraphs
  set noWordgraphs [llength $allWordgraphNames]
  set half [expr int($noWordgraphs * 0.5)]
  set selection ""
  expr srand([clock seconds])

  for {set i $half} {$i >=0} {incr i -1} {
    set index [expr int(rand() * $noWordgraphs)]
    lappend selection [lindex $allWordgraphNames $index]
    incr noWordgraphs -1
    set allWordgraphNames [lreplace $allWordgraphNames $index $index]
  }

  # select them
  _selectWordgraphs $selection
}

## ----------------------------------------------------------------------------
## select wordgraphs of a limited length from the current selection.
## If min or max has not been specified then a dialog pops up to get these
## values interactively.
## \see YadaGrammar::getWordgraphDistribution()
## \param min minimum length of a sentence to be selected (optional parameter)
## \param max maximum length of a sentence to be selected (optional parameter)
## \returns the selected wordgraph names
## ----------------------------------------------------------------------------
body YadaGrammars::_lengthSelection {{min -1} {max -1}} {

  set grammarName [$itk_component(name) get]
  if {$grammarName == "<none>"} {
    return
  }

  if {$min < 0 || $max < 0} {
    # TODO: selection dialog
    set min 0
    set max 60
  }

  # get current selected wordgraphs
  set allWordgraphNames [_getSelectedWordgraphs]

  # unselect all
  set table $itk_component(table)
  set rmax [$table cget -rows]
  for {set row 0} {$row < $rmax} {incr row} { 
    $table tag row "" $row
  }

  # construct a new selection
  set selection ""
  set grammar [getCurrentItem]
  foreach wordgraphName $allWordgraphNames {
    set wordgraph [$grammar getWordgraph $wordgraphName]
    if {$wordgraph == ""} {
      continue
    }
    set length [llength [$wordgraph getWords]]
    if {$length >= $min && $length <= $max} {
      lappend selection $wordgraphName
    }
  }

  # select them
  _selectWordgraphs $selection
}

## ----------------------------------------------------------------------------
## _getSelectedWordgraphs
## ----------------------------------------------------------------------------
body YadaGrammars::_getSelectedWordgraphs {} {
  set rmax [$itk_component(table) cget -rows]
  set result ""
  for {set row 1} {$row < $rmax} {incr row} {
    if {[$itk_component(table) tag includes mark $row,1]} {
      lappend result [$itk_component(table) getCell $row,1]
    }
  }

  return "$result"
}

## ----------------------------------------------------------------------------
## _rescanGrammar
## ----------------------------------------------------------------------------
body YadaGrammars::_rescanGrammar {} {
  set grammarName [$itk_component(name) get]
  if {$grammarName == "<none>"} {
    return
  }
   
  ## Set cursor busy and save old cursors
  set oldCursors [setBusy]
  set currentItem [getCurrentItem]
  $currentItem rescan
  _select
  updateWindowTitle
  updateMenus

  ## Restore old cursors
  resetBusy $oldCursors

  _unmark
}

## ----------------------------------------------------------------------------
## _fileChooser
## ----------------------------------------------------------------------------
body YadaGrammars::_fileChooser {} {
  set currentItem [getCurrentItem]
  if {$currentItem == ""} {
    return
  }

  set files [$currentItem cget -files]
  set directory [file dirname [lindex $files 0]]
  set selection [.fileSelector activate \
    -directory $directory  \
    -title "Choose CDG grammar files" \
    -filter {"*.cd? *.m4" "*.cd?" "*.m4" "*" } \
    -mode "multi" \
    -selection $files]

  if {$selection != ""} { 
    $currentItem configure -files $selection
    _select
  }
}

## ----------------------------------------------------------------------------
## Checks, whether given file contains valid grammar data, 
## returning an empty string, if it is not valid, otherwise
## the unique name of the grammar 
## ----------------------------------------------------------------------------
body YadaGrammars::isValidPersistenceFile {fileName} {

  if {![file exists $fileName]} {   
    #error "File $fileName not found."
    return ""
  }

  # open it
  set fileID [open "|zcat $fileName" {RDONLY}]
  set xmlData [read  $fileID]
  close $fileID
  set document [dom parse $xmlData]
  set rootNode [$document documentElement] 
  
  # and search the grammar
  set nodeGrammar [$rootNode selectNodes "/yada/grammar"]

  if {$nodeGrammar == ""} {
    return ""
  } else {
    return [$nodeGrammar getAttribute name]
  }
}

## ----------------------------------------------------------------------------
## insert stuff into a disabled table
## ----------------------------------------------------------------------------
body YadaGrammars::_setCell {args} {
  $itk_component(table) configure -state normal
  eval $itk_component(table) setCell $args
  $itk_component(table) configure -state disabled
}
